# Copyright 2016 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for asynchronous helpers."""

import asyncio
from inspect import isawaitable

from testtools.matchers import Equals, Is, MatchesPredicate

from .. import maas_async
from ...testing import TestCase


IsAwaitable = MatchesPredicate(isawaitable, "%s is not awaitable")


class TestAsynchronousWrapper(TestCase):
    """Tests for `asynchronous`."""

    def test_returns_plain_result_unaltered_when_loop_not_running(self):
        token = object()
        func = maas_async.asynchronous(lambda: token)
        self.assertThat(func(), Is(token))

    def test_returns_plain_result_unaltered_when_loop_running(self):
        token = object()
        func = maas_async.asynchronous(lambda: token)

        async def within_event_loop():
            loop = asyncio.get_event_loop()
            self.assertTrue(loop.is_running())
            return func()

        self.assertThat(self.loop.run_until_complete(within_event_loop()), Is(token))

    def test_blocks_on_awaitable_result_when_loop_not_running(self):
        token = asyncio.sleep(0.0)
        func = maas_async.asynchronous(lambda: token)
        self.assertThat(func(), Is(None))

    def test_returns_awaitable_result_unaltered_when_loop_running(self):
        token = asyncio.sleep(0.0)
        func = maas_async.asynchronous(lambda: token)

        async def within_event_loop():
            loop = asyncio.get_event_loop()
            self.assertTrue(loop.is_running())
            return func()

        result = self.loop.run_until_complete(within_event_loop())
        self.assertThat(result, Is(token))
        self.assertThat(result, IsAwaitable)
        result = self.loop.run_until_complete(result)
        self.assertThat(result, Is(None))


class TestAsynchronousType(TestCase):
    """Tests for `Asynchronous`."""

    def test_callable_attributes_are_wrapped(self):
        # `Asynchronous` groks class- and static-methods.

        class Class(metaclass=maas_async.Asynchronous):

            attribute = 123

            def imethod(self):
                return self, "instancemethod"

            @classmethod
            def cmethod(cls):
                return cls, "classmethod"

            @staticmethod
            def smethod():
                return None, "staticmethod"

        self.assertThat(Class.attribute, Equals(123))
        self.assertThat(Class.smethod(), Equals((None, "staticmethod")))
        self.assertThat(Class.cmethod(), Equals((Class, "classmethod")))

        inst = Class()
        self.assertThat(inst.attribute, Equals(123))
        self.assertThat(inst.smethod(), Equals((None, "staticmethod")))
        self.assertThat(inst.cmethod(), Equals((Class, "classmethod")))
        self.assertThat(inst.imethod(), Equals((inst, "instancemethod")))



"""Test for `maas.client.viscera.pods`."""

import random

from testtools.matchers import Equals, IsInstance

from ...errors import OperationNotAllowed
from ..pods import Pod, Pods
from ..testing import bind
from ...testing import make_name_without_spaces, make_string_without_spaces, TestCase


def make_origin():
    """
    Create a new origin with Pods and Pod. The former
    refers to the latter via the origin, hence why it must be bound.
    """
    return bind(Pods, Pod)


def make_pod():
    """Returns a pod dictionary."""
    return {
        "id": random.randint(1, 100),
        "type": make_name_without_spaces("type"),
        "name": make_name_without_spaces("name"),
        "architectures": [make_string_without_spaces() for _ in range(3)],
        "capabilities": [make_string_without_spaces() for _ in range(3)],
        "zone": {
            "id": random.randint(1, 100),
            "name": make_name_without_spaces("name"),
            "description": make_name_without_spaces("description"),
        },
        "tags": [make_string_without_spaces() for _ in range(3)],
        "cpu_over_commit_ratio": random.uniform(0, 10),
        "memory_over_commit_ratio": random.uniform(0, 10),
        "available": {
            "cores": random.randint(1, 100),
            "memory": random.randint(4096, 8192),
            "local_storage": random.randint(1024, 1024 * 1024),
        },
        "used": {
            "cores": random.randint(1, 100),
            "memory": random.randint(4096, 8192),
            "local_storage": random.randint(1024, 1024 * 1024),
        },
        "total": {
            "cores": random.randint(1, 100),
            "memory": random.randint(4096, 8192),
            "local_storage": random.randint(1024, 1024 * 1024),
        },
    }


class TestPods(TestCase):
    def test__pods_create(self):
        type = make_string_without_spaces()
        power_address = make_string_without_spaces()
        power_user = make_string_without_spaces()
        power_pass = make_string_without_spaces()
        name = make_string_without_spaces()
        zone = make_string_without_spaces()
        tags = make_string_without_spaces()
        origin = make_origin()
        Pods, Pod = origin.Pods, origin.Pod
        Pods._handler.create.return_value = {}
        observed = Pods.create(
            type=type,
            power_address=power_address,
            power_user=power_user,
            power_pass=power_pass,
            name=name,
            zone=zone,
            tags=tags,
        )
        self.assertThat(observed, IsInstance(Pod))
        Pods._handler.create.assert_called_once_with(
            type=type,
            power_address=power_address,
            power_user=power_user,
            power_pass=power_pass,
            name=name,
            zone=zone,
            tags=tags,
        )

    def test__pods_create_raises_error_for_rsd_and_no_power_user(self):
        origin = make_origin()
        origin.Pods._handler.create.return_value = {}
        self.assertRaises(
            OperationNotAllowed,
            origin.Pods.create,
            type="rsd",
            power_address=make_string_without_spaces(),
        )

    def test__pods_create_raises_error_for_rsd_and_no_power_pass(self):
        origin = make_origin()
        origin.Pods._handler.create.return_value = {}
        self.assertRaises(
            OperationNotAllowed,
            origin.Pods.create,
            type="rsd",
            power_address=make_string_without_spaces(),
            power_user=make_string_without_spaces(),
        )

    def test__pods_create_raises_type_error_for_zone(self):
        origin = make_origin()
        origin.Pods._handler.create.return_value = {}
        self.assertRaises(
            TypeError,
            origin.Pods.create,
            type=make_string_without_spaces(),
            power_address=make_string_without_spaces(),
            power_user=make_string_without_spaces(),
            power_pass=make_string_without_spaces(),
            zone=0.1,
        )

    def test__pods_read(self):
        Pods = make_origin().Pods
        pods = [make_pod() for _ in range(3)]
        Pods._handler.read.return_value = pods
        pods = Pods.read()
        self.assertThat(len(pods), Equals(3))


class TestPod(TestCase):
    def test__pod_read(self):
        Pod = make_origin().Pod
        pod = make_pod()
        Pod._handler.read.return_value = pod
        self.assertThat(Pod.read(id=pod["id"]), Equals(Pod(pod)))
        Pod._handler.read.assert_called_once_with(id=pod["id"])

    def test__pod_refresh(self):
        Pod = make_origin().Pod
        pod_data = make_pod()
        pod = Pod(pod_data)
        pod.refresh()
        Pod._handler.refresh.assert_called_once_with(id=pod_data["id"])

    def test__pod_parameters(self):
        Pod = make_origin().Pod
        pod_data = make_pod()
        pod = Pod(pod_data)
        pod.parameters()
        Pod._handler.parameters.assert_called_once_with(id=pod_data["id"])

    def test__pod_compose(self):
        Pod = make_origin().Pod
        pod_data = make_pod()
        pod = Pod(pod_data)
        cores = random.randint(1, 100)
        memory = random.randint(4096, 8192)
        cpu_speed = random.randint(16, 256)
        architecture = make_name_without_spaces("architecture")
        storage = make_string_without_spaces()
        hostname = make_name_without_spaces("hostname")
        domain = random.randint(1, 10)
        zone = random.randint(1, 10)
        interfaces = make_string_without_spaces()
        pod.compose(
            cores=cores,
            memory=memory,
            cpu_speed=cpu_speed,
            architecture=architecture,
            storage=storage,
            hostname=hostname,
            domain=domain,
            zone=zone,
            interfaces=interfaces,
        )
        Pod._handler.compose.assert_called_once_with(
            id=pod_data["id"],
            cores=str(cores),
            memory=str(memory),
            cpu_speed=str(cpu_speed),
            architecture=architecture,
            storage=storage,
            hostname=hostname,
            domain=str(domain),
            zone=str(zone),
            interfaces=interfaces,
        )

    def test__pod_compose_raises_type_error_for_zone(self):
        Pod = make_origin().Pod
        pod_data = make_pod()
        pod = Pod(pod_data)
        cores = random.randint(1, 100)
        memory = random.randint(4096, 8192)
        cpu_speed = random.randint(16, 256)
        architecture = make_name_without_spaces("architecture")
        storage = make_string_without_spaces()
        hostname = make_name_without_spaces("hostname")
        domain = random.randint(1, 10)
        zone = 0.1
        interfaces = make_string_without_spaces()
        self.assertRaises(
            TypeError,
            pod.compose,
            cores=cores,
            memory=memory,
            cpu_speed=cpu_speed,
            architecture=architecture,
            storage=storage,
            hostname=hostname,
            domain=domain,
            zone=zone,
            interfaces=interfaces,
        )

    def test__pod_delete(self):
        Pod = make_origin().Pod
        pod_data = make_pod()
        pod = Pod(pod_data)
        pod.delete()
        Pod._handler.delete.assert_called_once_with(id=pod_data["id"])

    def test__save_add_tag(self):
        Pod = make_origin().Pod
        pod_data = make_pod()
        pod = Pod(pod_data)
        tag = make_string_without_spaces()
        pod.tags.append(tag)
        Pod._handler.add_tag.return_value = None
        pod.save()
        Pod._handler.add_tag.assert_called_once_with(id=pod.id, tag=tag)

    def test__save_remove_tag(self):
        Pod = make_origin().Pod
        pod_data = make_pod()
        tag = make_string_without_spaces()
        pod_data["tags"] = [tag]
        pod = Pod(pod_data)
        pod.tags.remove(tag)
        Pod._handler.remove_tag.return_value = None
        pod.save()
        Pod._handler.remove_tag.assert_called_once_with(id=pod.id, tag=tag)



"""Commands for subnets."""

__all__ = ["register"]

from http import HTTPStatus

from . import CommandError, OriginPagedTableCommand, tables
from ..bones import CallError
from ..utils.maas_async import asynchronous


class cmd_subnets(OriginPagedTableCommand):
    """List subnets."""

    def __init__(self, parser):
        super(cmd_subnets, self).__init__(parser)
        parser.add_argument(
            "--minimal", action="store_true", help=("Output only the subnet names.")
        )

    @asynchronous
    async def load_object_sets(self, origin):
        subnets = origin.Subnets.read()
        fabrics = origin.Fabrics.read()
        return await subnets, await fabrics

    def execute(self, origin, options, target):
        visible_columns = None
        if options.minimal:
            visible_columns = ("name",)
        subnets, fabrics = self.load_object_sets(origin)
        table = tables.SubnetsTable(visible_columns=visible_columns, fabrics=fabrics)
        return table.render(target, subnets)


class cmd_subnet(OriginPagedTableCommand):
    """Details of a subnet."""

    def __init__(self, parser):
        super(cmd_subnet, self).__init__(parser)
        parser.add_argument("name", nargs=1, help=("Name of the subnet."))

    def execute(self, origin, options, target):
        try:
            subnet = origin.Subnet.read(options.name[0])
        except CallError as error:
            if error.status == HTTPStatus.NOT_FOUND:
                raise CommandError("Unable to find subnet %s." % options.name[0])
            else:
                raise
        table = tables.SubnetDetail(fabrics=origin.Fabrics.read())
        return table.render(target, subnet)


def register(parser):
    """Register commands with the given parser."""
    cmd_subnets.register(parser)
    cmd_subnet.register(parser)



"""Test for `maas.client.viscera.vlans`."""

import random
from operator import itemgetter

from testtools.matchers import Equals

from ..controllers import RackController
from ..fabrics import Fabric, Fabrics
from ..vlans import Vlan, Vlans

from ..testing import bind
from ...testing import TestCase


def make_origin():
    """
    Create a new origin with Vlans, Vlan, Fabric and RackController. The former
    refers to the latter via the origin, hence why it must be bound.
    """
    return bind(Vlans, Vlan, Fabric, Fabrics, RackController)


class TestVlans(TestCase):
    def test__vlans_create(self):
        Vlans = make_origin().Vlans
        fabric_id = random.randint(1, 100)
        vid = random.randint(1, 100)
        Vlans._handler.create.return_value = {
            "id": 5001,
            "fabric_id": fabric_id,
            "vid": vid,
            "mtu": 1500,
        }
        Vlans.create(fabric_id, vid)
        Vlans._handler.create.assert_called_once_with(
            fabric_id=fabric_id, vid=vid, dhcp_on=False
        )

    def test__vlans_create_with_fabric(self):
        Vlans = make_origin().Vlans
        Fabric = make_origin().Fabric
        fabric_id = random.randint(1, 100)
        fabric = Fabric({"id": fabric_id})
        vid = random.randint(1, 100)
        Vlans._handler.create.return_value = {
            "id": 5001,
            "fabric_id": fabric_id,
            "vid": vid,
            "mtu": 1500,
        }
        Vlans.create(fabric, vid)
        Vlans._handler.create.assert_called_once_with(
            fabric_id=fabric_id, vid=vid, dhcp_on=False
        )

    def test__vlans_read(self):
        Vlans = make_origin().Vlans
        fabric_id = random.randint(1, 100)
        vlans = [
            {"id": random.randint(5001, 6000), "vid": random.randint(1, 4096)}
            for _ in range(3)
        ]
        Vlans._handler.read.return_value = vlans
        vlans = Vlans.read(fabric_id)
        self.assertThat(len(vlans), Equals(3))

    def test__vlans_read_with_fabric(self):
        Vlans = make_origin().Vlans
        Fabric = make_origin().Fabric
        fabric_id = random.randint(1, 100)
        fabric = Fabric({"id": fabric_id})
        vlans = [
            {"id": random.randint(5001, 6000), "vid": random.randint(1, 4096)}
            for _ in range(3)
        ]
        Vlans._handler.read.return_value = vlans
        vlans = Vlans.read(fabric)
        self.assertThat(len(vlans), Equals(3))

    def test__vlans_get_default(self):
        Vlans = make_origin().Vlans
        Fabric = make_origin().Fabric
        fabric_id = random.randint(1, 100)
        fabric = Fabric({"id": fabric_id})
        vlans = [
            {"id": random.randint(5001, 6000), "vid": random.randint(1, 4096)}
            for _ in range(3)
        ]
        default_vlan = sorted(vlans, key=itemgetter("id"))[0]
        Vlans._handler.read.return_value = vlans
        vlans = Vlans.read(fabric)
        observed = vlans.get_default()
        self.assertThat(observed.id, Equals(default_vlan["id"]))


class TestVlan(TestCase):
    def test__vlan_read(self):
        Vlan = make_origin().Vlan
        Vlan._handler.read.return_value = {"id": 5001, "vid": 10}
        fabric_id = random.randint(1, 100)
        Vlan.read(fabric_id, 10)
        Vlan._handler.read.assert_called_once_with(fabric_id=fabric_id, vid=10)

    def test__vlan_delete(self):
        Vlan = make_origin().Vlan
        fabric_id = random.randint(1, 100)
        vlan = Vlan({"id": 5001, "fabric_id": fabric_id, "vid": 10})
        vlan.delete()
        Vlan._handler.delete.assert_called_once_with(fabric_id=fabric_id, vid=10)

    def test__vlan_update_fabric_to_default(self):
        origin = make_origin()
        Fabric, Vlan = origin.Fabric, origin.Vlan
        Vlan._handler.params = ["fabric_id", "vid"]
        fabric_id = random.randint(1, 100)
        vlan_id = random.randint(5001, 6000)
        vid = random.randint(100, 200)
        vlan = Vlan({"id": vlan_id, "fabric_id": fabric_id, "vid": vid})
        default_fabric = Fabric({"id": 0})
        vlan.fabric = default_fabric
        Vlan._handler.update.return_value = {"id": vlan_id, "vid": vid}
        vlan.save()
        Vlan._handler.update.assert_called_once_with(
            fabric_id=fabric_id, vid=vid, _vid=vid, fabric=default_fabric.id
        )
        self.assertThat(vlan.fabric.id, Equals(default_fabric.id))

    def test__vlan_update_fabric(self):
        origin = make_origin()
        Fabric, Vlan = origin.Fabric, origin.Vlan
        Vlan._handler.params = ["fabric_id", "vid"]
        fabric_id = random.randint(1, 100)
        vlan_id = random.randint(5001, 6000)
        vid = random.randint(100, 200)
        vlan = Vlan({"id": vlan_id, "fabric_id": fabric_id, "vid": vid})
        new_fabric = Fabric({"id": random.randint(101, 200)})
        vlan.fabric = new_fabric
        Vlan._handler.update.return_value = {"id": vlan_id, "vid": vid}
        vlan.save()
        Vlan._handler.update.assert_called_once_with(
            fabric_id=fabric_id, vid=vid, _vid=vid, fabric=new_fabric.id
        )
        self.assertThat(vlan.fabric.id, Equals(new_fabric.id))

    def test__vlan_update_vid(self):
        origin = make_origin()
        Vlan = origin.Vlan
        Vlan._handler.params = ["fabric_id", "vid"]
        fabric_id = random.randint(1, 100)
        vlan_id = random.randint(5001, 6000)
        vid = random.randint(100, 200)
        new_vid = random.randint(201, 300)
        vlan = Vlan({"id": vlan_id, "fabric_id": fabric_id, "vid": vid})
        vlan.vid = new_vid
        Vlan._handler.update.return_value = {"id": vlan_id, "vid": new_vid}
        vlan.save()
        Vlan._handler.update.assert_called_once_with(
            fabric_id=fabric_id, vid=vid, _vid=new_vid
        )
        self.assertThat(new_vid, Equals(new_vid))

    def test__vlan_update_vid_twice(self):
        origin = make_origin()
        Vlan = origin.Vlan
        Vlan._handler.params = ["fabric_id", "vid"]
        fabric_id = random.randint(1, 100)
        vlan_id = random.randint(5001, 6000)
        vid = random.randint(100, 200)
        new_vid = random.randint(201, 300)
        vlan = Vlan({"id": vlan_id, "fabric_id": fabric_id, "vid": vid})
        vlan.vid = new_vid
        Vlan._handler.update.return_value = {"id": vlan_id, "vid": new_vid}
        vlan.save()
        Vlan._handler.update.assert_called_once_with(
            fabric_id=fabric_id, vid=vid, _vid=new_vid
        )
        self.assertThat(vlan.vid, Equals(new_vid))

        # Second call should pass the new_vid as the vid, and not the old vid.
        new_vid_2 = random.randint(301, 400)
        vlan.vid = new_vid_2
        Vlan._handler.update.return_value = {"id": vlan_id, "vid": new_vid_2}
        vlan.save()
        Vlan._handler.update.assert_called_with(
            fabric_id=fabric_id, vid=new_vid, _vid=new_vid_2
        )
        self.assertThat(vlan.vid, Equals(new_vid_2))

    def test__vlan_update_relay_vlan_with_object(self):
        origin = make_origin()
        Vlan = origin.Vlan
        Vlan._handler.params = ["fabric_id", "vid"]
        fabric_id = random.randint(1, 100)
        vlan_id = random.randint(5001, 6000)
        vid = random.randint(100, 200)
        vlan = Vlan(
            {
                "id": vlan_id,
                "fabric_id": fabric_id,
                "vid": vid,
                "name": "",
                "relay_vlan": None,
            }
        )
        relay_vlan = Vlan(
            {"id": vlan_id + 1, "fabric_id": fabric_id, "vid": vid + 10, "name": ""}
        )
        vlan.relay_vlan = relay_vlan
        Vlan._handler.update.return_value = {
            "id": vlan_id,
            "vid": vid,
            "name": "",
            "relay_vlan": relay_vlan._data,
        }
        vlan.save()
        Vlan._handler.update.assert_called_once_with(
            fabric_id=fabric_id, vid=vid, _vid=vid, relay_vlan=relay_vlan.id
        )
        self.assertThat(vlan.relay_vlan.id, Equals(relay_vlan.id))

    def test__vlan_update_relay_vlan_with_integer_id(self):
        self.skip("see https://github.com/maas/python-libmaas/issues/180")
        origin = make_origin()
        Vlan = origin.Vlan
        Vlan._handler.params = ["fabric_id", "vid"]
        fabric_id = random.randint(1, 100)
        vlan_id = random.randint(5001, 6000)
        vid = random.randint(100, 200)
        vlan = Vlan(
            {"id": vlan_id, "fabric_id": fabric_id, "vid": vid, "relay_vlan": None}
        )
        relay_vlan = Vlan({"id": vlan_id + 1, "fabric_id": fabric_id, "vid": vid + 10})
        vlan.relay_vlan = str(vlan_id + 1)
        Vlan._handler.update.return_value = {
            "id": vlan_id,
            "vid": vid,
            "name": "",
            "relay_vlan": relay_vlan._data,
        }
        vlan.save()
        Vlan._handler.update.assert_called_once_with(
            fabric_id=fabric_id, vid=vid, _vid=vid, relay_vlan=relay_vlan.id
        )
        self.assertThat(vlan.relay_vlan.id, Equals(relay_vlan.id))



"""Test for `maas.client.viscera.nodes`."""

from copy import deepcopy

from testtools.matchers import (
    Equals,
    IsInstance,
    MatchesAll,
    MatchesListwise,
    MatchesStructure,
)

from .. import nodes
from ..controllers import RackController, RegionController
from ..tags import Tags, Tag
from ..devices import Device
from ..domains import Domain
from ..machines import Machine
from ..resource_pools import ResourcePool
from ..testing import bind
from ...enum import NodeType
from ...testing import make_name_without_spaces, TestCase


def make_origin():
    # Create a new origin with Nodes and Node. The former refers to the
    # latter via the origin, hence why it must be bound.
    return bind(
        nodes.Nodes,
        nodes.Node,
        Device,
        Domain,
        Machine,
        RackController,
        RegionController,
        ResourcePool,
        Tags,
        Tag,
    )


class TestNode(TestCase):
    def test__string_representation_includes_only_system_id_and_hostname(self):
        node = nodes.Node(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
            }
        )
        self.assertThat(
            repr(node),
            Equals("<Node hostname=%(hostname)r system_id=%(system_id)r>" % node._data),
        )

    def test__read(self):
        data = {
            "system_id": make_name_without_spaces("system-id"),
            "hostname": make_name_without_spaces("hostname"),
        }

        origin = make_origin()
        origin.Node._handler.read.return_value = data

        node_observed = origin.Node.read(data["system_id"])
        node_expected = origin.Node(data)
        self.assertThat(node_observed, Equals(node_expected))

    def test__read_domain(self):
        domain = {"name": make_name_without_spaces("domain")}
        data = {
            "system_id": make_name_without_spaces("system-id"),
            "hostname": make_name_without_spaces("hostname"),
            "domain": domain,
        }
        origin = make_origin()
        origin.Node._handler.read.return_value = data
        node = origin.Node.read(data["system_id"])
        domain = origin.Domain(domain)
        self.assertThat(node.domain, Equals(domain))

    def test__save_change_pool(self):
        pool_data = {"id": 1, "name": "pool1", "description": "pool1"}
        new_pool_data = {"id": 2, "name": "pool2", "description": "pool2"}
        system_id = make_name_without_spaces("system-id")
        node_data = {
            "id": 1,
            "system_id": system_id,
            "hostname": make_name_without_spaces("hostname"),
            "pool": pool_data,
        }

        origin = make_origin()
        origin.ResourcePool._handler.read.return_value = new_pool_data
        origin.Node._handler.params = ["system_id", "id"]
        origin.Node._handler.read.return_value = node_data
        origin.Node._handler.update.return_value = deepcopy(node_data)
        origin.Node._handler.update.return_value["pool"] = new_pool_data

        new_pool = origin.ResourcePool.read(2)
        node = origin.Node.read(system_id)
        node.pool = new_pool
        node.save()
        origin.Node._handler.update.assert_called_once_with(
            id=1, pool="pool2", system_id=system_id
        )

    def test__save_change_domain(self):
        domain_data = {"id": 1, "name": "domain1"}
        new_domain_data = {"id": 2, "name": "domain2"}
        system_id = make_name_without_spaces("system-id")
        node_data = {
            "id": 1,
            "system_id": system_id,
            "hostname": make_name_without_spaces("hostname"),
            "domain": domain_data,
        }

        origin = make_origin()
        origin.Domain._handler.read.return_value = new_domain_data
        origin.Node._handler.params = ["system_id", "id"]
        origin.Node._handler.read.return_value = node_data
        origin.Node._handler.update.return_value = deepcopy(node_data)
        origin.Node._handler.update.return_value["domain"] = new_domain_data

        new_domain = origin.Domain.read(2)
        node = origin.Node.read(system_id)
        node.domain = new_domain
        node.save()
        origin.Node._handler.update.assert_called_once_with(
            id=1, domain=2, system_id=system_id
        )

    def test__as_machine_requires_machine_type(self):
        origin = make_origin()
        device_node = origin.Node(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
                "node_type": NodeType.DEVICE.value,
            }
        )
        rack_node = origin.Node(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
                "node_type": NodeType.RACK_CONTROLLER.value,
            }
        )
        region_node = origin.Node(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
                "node_type": NodeType.REGION_CONTROLLER.value,
            }
        )
        region_rack_node = origin.Node(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
                "node_type": NodeType.REGION_AND_RACK_CONTROLLER.value,
            }
        )
        self.assertRaises(ValueError, device_node.as_machine)
        self.assertRaises(ValueError, rack_node.as_machine)
        self.assertRaises(ValueError, region_node.as_machine)
        self.assertRaises(ValueError, region_rack_node.as_machine)

    def test__as_machine_returns_machine_type(self):
        origin = make_origin()
        machine_node = origin.Node(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
                "node_type": NodeType.MACHINE.value,
            }
        )
        machine = machine_node.as_machine()
        self.assertIsInstance(machine, Machine)

    def test__as_device_requires_device_type(self):
        origin = make_origin()
        machine_node = origin.Node(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
                "node_type": NodeType.MACHINE.value,
            }
        )
        rack_node = origin.Node(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
                "node_type": NodeType.RACK_CONTROLLER.value,
            }
        )
        region_node = origin.Node(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
                "node_type": NodeType.REGION_CONTROLLER.value,
            }
        )
        region_rack_node = origin.Node(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
                "node_type": NodeType.REGION_AND_RACK_CONTROLLER.value,
            }
        )
        self.assertRaises(ValueError, machine_node.as_device)
        self.assertRaises(ValueError, rack_node.as_device)
        self.assertRaises(ValueError, region_node.as_device)
        self.assertRaises(ValueError, region_rack_node.as_device)

    def test__as_device_returns_device_type(self):
        origin = make_origin()
        device_node = origin.Node(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
                "node_type": NodeType.DEVICE.value,
            }
        )
        device = device_node.as_device()
        self.assertIsInstance(device, Device)

    def test__as_rack_controller_requires_rack_types(self):
        origin = make_origin()
        machine_node = origin.Node(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
                "node_type": NodeType.MACHINE.value,
            }
        )
        device_node = origin.Node(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
                "node_type": NodeType.DEVICE.value,
            }
        )
        region_node = origin.Node(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
                "node_type": NodeType.REGION_CONTROLLER.value,
            }
        )
        self.assertRaises(ValueError, machine_node.as_rack_controller)
        self.assertRaises(ValueError, device_node.as_rack_controller)
        self.assertRaises(ValueError, region_node.as_rack_controller)

    def test__as_rack_controller_returns_rack_type(self):
        origin = make_origin()
        rack_node = origin.Node(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
                "node_type": NodeType.RACK_CONTROLLER.value,
            }
        )
        region_rack_node = origin.Node(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
                "node_type": NodeType.REGION_AND_RACK_CONTROLLER.value,
            }
        )
        rack = rack_node.as_rack_controller()
        region_rack = region_rack_node.as_rack_controller()
        self.assertIsInstance(rack, RackController)
        self.assertIsInstance(region_rack, RackController)

    def test__as_region_controller_requires_region_types(self):
        origin = make_origin()
        machine_node = origin.Node(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
                "node_type": NodeType.MACHINE.value,
            }
        )
        device_node = origin.Node(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
                "node_type": NodeType.DEVICE.value,
            }
        )
        rack_node = origin.Node(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
                "node_type": NodeType.RACK_CONTROLLER.value,
            }
        )
        self.assertRaises(ValueError, machine_node.as_region_controller)
        self.assertRaises(ValueError, device_node.as_region_controller)
        self.assertRaises(ValueError, rack_node.as_region_controller)

    def test__as_region_controller_returns_region_type(self):
        origin = make_origin()
        region_node = origin.Node(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
                "node_type": NodeType.REGION_CONTROLLER.value,
            }
        )
        region_rack_node = origin.Node(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
                "node_type": NodeType.REGION_AND_RACK_CONTROLLER.value,
            }
        )
        region = region_node.as_region_controller()
        region_rack = region_rack_node.as_region_controller()
        self.assertIsInstance(region, RegionController)
        self.assertIsInstance(region_rack, RegionController)

    def test__delete(self):
        Node = make_origin().Node

        system_id = make_name_without_spaces("system-id")
        node = Node(
            {
                "id": 1,
                "system_id": system_id,
                "hostname": make_name_without_spaces("hostname"),
            }
        )

        node.delete()
        Node._handler.delete.assert_called_once_with(system_id=node.system_id)

    def test__tags(self):
        origin = make_origin()

        tag_names = [make_name_without_spaces("tag") for _ in range(3)]

        system_id = make_name_without_spaces("system-id")
        node = origin.Node(
            {
                "id": 1,
                "system_id": system_id,
                "hostname": make_name_without_spaces("hostname"),
                "tag_names": tag_names,
            }
        )

        self.assertThat(
            node.tags,
            MatchesListwise(
                [
                    MatchesAll(
                        IsInstance(origin.Tag), MatchesStructure(name=Equals(tag_name))
                    )
                    for tag_name in tag_names
                ]
            ),
        )

    def test__tags_add(self):
        origin = make_origin()

        tag_names = [make_name_without_spaces("tag") for _ in range(3)]

        system_id = make_name_without_spaces("system-id")
        node = origin.Node(
            {
                "id": 1,
                "system_id": system_id,
                "hostname": make_name_without_spaces("hostname"),
                "tag_names": list(tag_names),
            }
        )

        tag = origin.Tag({"name": make_name_without_spaces("newtag")})
        node.tags.add(tag)

        origin.Tag._handler.update_nodes.assert_called_once_with(
            name=tag.name, add=node.system_id
        )

        self.assertThat(
            node.tags,
            MatchesListwise(
                [
                    MatchesAll(
                        IsInstance(origin.Tag), MatchesStructure(name=Equals(tag_name))
                    )
                    for tag_name in tag_names + [tag.name]
                ]
            ),
        )

    def test__tags_add_requires_Tag(self):
        origin = make_origin()

        tag_names = [make_name_without_spaces("tag") for _ in range(3)]

        system_id = make_name_without_spaces("system-id")
        node = origin.Node(
            {
                "id": 1,
                "system_id": system_id,
                "hostname": make_name_without_spaces("hostname"),
                "tag_names": list(tag_names),
            }
        )

        tag = origin.Tag({"name": make_name_without_spaces("newtag")})
        self.assertRaises(TypeError, node.tags.add, tag.name)

    def test__tags_remove(self):
        origin = make_origin()

        tag_names = [make_name_without_spaces("tag") for _ in range(3)]

        system_id = make_name_without_spaces("system-id")
        node = origin.Node(
            {
                "id": 1,
                "system_id": system_id,
                "hostname": make_name_without_spaces("hostname"),
                "tag_names": list(tag_names),
            }
        )

        tag = origin.Tag({"name": tag_names[0]})
        node.tags.remove(tag)

        origin.Tag._handler.update_nodes.assert_called_once_with(
            name=tag.name, remove=node.system_id
        )

        self.assertThat(
            node.tags,
            MatchesListwise(
                [
                    MatchesAll(
                        IsInstance(origin.Tag), MatchesStructure(name=Equals(tag_name))
                    )
                    for tag_name in tag_names[1:]
                ]
            ),
        )

    def test__tags_remove_requires_Tag(self):
        origin = make_origin()

        tag_names = [make_name_without_spaces("tag") for _ in range(3)]

        system_id = make_name_without_spaces("system-id")
        node = origin.Node(
            {
                "id": 1,
                "system_id": system_id,
                "hostname": make_name_without_spaces("hostname"),
                "tag_names": list(tag_names),
            }
        )

        tag = origin.Tag({"name": tag_names[0]})
        self.assertRaises(TypeError, node.tags.remove, tag.name)


class TestNodes(TestCase):
    def test__read(self):
        data = {
            "system_id": make_name_without_spaces("system-id"),
            "hostname": make_name_without_spaces("hostname"),
        }

        origin = make_origin()
        origin.Nodes._handler.read.return_value = [data]

        nodes_observed = origin.Nodes.read()
        nodes_expected = origin.Nodes([origin.Node(data)])
        self.assertThat(nodes_observed, Equals(nodes_expected))

    def test__read_with_hostnames(self):
        origin = make_origin()
        origin.Nodes._handler.read.return_value = []

        hostnames = [make_name_without_spaces() for _ in range(3)]
        origin.Nodes.read(hostnames=hostnames)
        origin.Nodes._handler.read.assert_called_once_with(hostname=hostnames)

    def test__read_with_normalized_hostnames(self):
        origin = make_origin()
        origin.Nodes._handler.read.return_value = []

        hostnames = [make_name_without_spaces() for _ in range(3)]
        origin.Nodes.read(
            hostnames=[
                "%s.%s" % (hostname, make_name_without_spaces())
                for hostname in hostnames
            ]
        )
        origin.Nodes._handler.read.assert_called_once_with(hostname=hostnames)



"""Tests for `maas.client.flesh.machines`."""

from functools import partial
from operator import itemgetter
import yaml

from .testing import TestCaseWithProfile
from .. import ArgumentParser, machines, tabular
from ...enum import NodeStatus, PowerState
from ...testing import make_name_without_spaces
from ...viscera.testing import bind
from ...viscera.machines import Machine, Machines
from ...viscera.resource_pools import ResourcePool
from ...viscera.tags import Tag, Tags
from ...viscera.users import User
from ...viscera.zones import Zone


def make_origin():
    """Make origin for machines."""
    return bind(Machines, Machine, User, ResourcePool, Zone, Tag, Tags)


class TestMachines(TestCaseWithProfile):
    """Tests for `cmd_machines`."""

    def test_returns_table_with_machines(self):
        origin = make_origin()
        parser = ArgumentParser()
        machine_objs = [
            {
                "hostname": make_name_without_spaces(),
                "architecture": "amd64/generic",
                "status": NodeStatus.READY.value,
                "status_name": NodeStatus.READY.name,
                "owner": None,
                "power_state": PowerState.OFF.value,
                "cpu_count": 2,
                "memory": 1024,
                "pool": {"id": 1, "name": "pool1", "description": "pool1"},
                "zone": {"id": 1, "name": "zone1", "description": "zone1"},
            },
            {
                "hostname": make_name_without_spaces(),
                "architecture": "i386/generic",
                "status": NodeStatus.DEPLOYED.value,
                "status_name": NodeStatus.DEPLOYED.name,
                "owner": make_name_without_spaces(),
                "power_state": PowerState.ON.value,
                "cpu_count": 4,
                "memory": 4096,
                "pool": {"id": 2, "name": "pool2", "description": "pool2"},
                "zone": {"id": 2, "name": "zone2", "description": "zone2"},
            },
        ]
        origin.Machines._handler.read.return_value = machine_objs
        cmd = machines.cmd_machines(parser)
        subparser = machines.cmd_machines.register(parser)
        options = subparser.parse_args([])
        output = yaml.safe_load(
            cmd.execute(origin, options, target=tabular.RenderTarget.yaml)
        )
        self.assertEquals(
            [
                {"name": "hostname", "title": "Hostname"},
                {"name": "power", "title": "Power"},
                {"name": "status", "title": "Status"},
                {"name": "owner", "title": "Owner"},
                {"name": "architecture", "title": "Arch"},
                {"name": "cpus", "title": "#CPUs"},
                {"name": "memory", "title": "RAM"},
                {"name": "pool", "title": "Resource pool"},
                {"name": "zone", "title": "Zone"},
            ],
            output["columns"],
        )
        machines_output = sorted(
            [
                {
                    "hostname": machine["hostname"],
                    "power": machine["power_state"],
                    "status": machine["status_name"],
                    "owner": machine["owner"] if machine["owner"] else "(none)",
                    "architecture": machine["architecture"],
                    "cpus": machine["cpu_count"],
                    "memory": machine["memory"],
                    "pool": machine["pool"]["name"],
                    "zone": machine["zone"]["name"],
                }
                for machine in machine_objs
            ],
            key=itemgetter("hostname"),
        )
        self.assertEquals(machines_output, output["data"])

    def test_calls_handler_with_hostnames(self):
        origin = make_origin()
        parser = ArgumentParser()
        origin.Machines._handler.read.return_value = []
        subparser = machines.cmd_machines.register(parser)
        cmd = machines.cmd_machines(parser)
        hostnames = [make_name_without_spaces() for _ in range(3)]
        options = subparser.parse_args(hostnames)
        cmd.execute(origin, options, target=tabular.RenderTarget.yaml)
        origin.Machines._handler.read.assert_called_once_with(hostname=hostnames)


class TestMachine(TestCaseWithProfile):
    """Tests for `cmd_machine`."""

    def setUp(self):
        super().setUp()
        origin = make_origin()
        parser = ArgumentParser()
        self.hostname = make_name_without_spaces()
        machine_objs = [
            {
                "hostname": self.hostname,
                "architecture": "amd64/generic",
                "status": NodeStatus.READY.value,
                "status_name": NodeStatus.READY.name,
                "owner": None,
                "power_state": PowerState.OFF.value,
                "cpu_count": 2,
                "memory": 1024,
                "pool": {"id": 1, "name": "pool1", "description": "pool1"},
                "zone": {"id": 1, "name": "zone1", "description": "zone1"},
                "tag_names": ["tag1", "tag2"],
                "distro_series": "",
                "power_type": "Manual",
            },
        ]
        origin.Machines._handler.read.return_value = machine_objs
        cmd = machines.cmd_machine(parser)
        subparser = machines.cmd_machine.register(parser)
        options = subparser.parse_args([machine_objs[0]["hostname"]])
        self.cmd = partial(cmd.execute, origin, options)

    def test_yaml_machine_details_with_tags(self):
        yaml_output = yaml.safe_load(self.cmd(target=tabular.RenderTarget.yaml))
        self.assertEqual(yaml_output.get("tags"), ["tag1", "tag2"])

    def test_plain_machine_details_with_tags(self):
        plain_output = self.cmd(target=tabular.RenderTarget.plain)
        self.assertEqual(
            plain_output,
            f"""\
+---------------+-------------+
| Hostname      | {self.hostname} |
| Status        | READY       |
| Image         | (none)      |
| Power         | Off         |
| Power Type    | Manual      |
| Arch          | amd64       |
| #CPUs         | 2           |
| RAM           | 1.0 GB      |
| Interfaces    | 0 physical  |
| IP addresses  |             |
| Resource pool | pool1       |
| Zone          | zone1       |
| Owner         | (none)      |
| Tags          | tag1        |
|               | tag2        |
+---------------+-------------+""",
        )



# Copyright 2017 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helpers for calulating difference between objects."""

__all__ = ["calculate_dict_diff"]


from . import remove_None


def calculate_dict_diff(old_params: dict, new_params: dict):
    """Return the parameters based on the difference.

    If a parameter exists in `old_params` but not in `new_params` then
    parameter will be set to an empty string.
    """
    # Ignore all None values as those cannot be saved.
    old_params = remove_None(old_params)
    new_params = remove_None(new_params)
    params_diff = {}
    for key, value in old_params.items():
        if key in new_params:
            if value != new_params[key]:
                params_diff[key] = new_params[key]
        else:
            params_diff[key] = ""
    for key, value in new_params.items():
        if key not in old_params:
            params_diff[key] = value
    return params_diff



"""Commands for fabrics."""

__all__ = ["register"]

from http import HTTPStatus

from . import CommandError, OriginPagedTableCommand, tables
from ..bones import CallError
from ..utils.maas_async import asynchronous


class cmd_fabrics(OriginPagedTableCommand):
    """List fabrics."""

    def __init__(self, parser):
        super(cmd_fabrics, self).__init__(parser)
        parser.add_argument(
            "--minimal", action="store_true", help=("Output only the fabric names.")
        )

    @asynchronous
    async def load_object_sets(self, origin):
        fabrics = origin.Fabrics.read()
        subnets = origin.Subnets.read()
        return await fabrics, await subnets

    def execute(self, origin, options, target):
        visible_columns = None
        if options.minimal:
            visible_columns = ("name",)
        fabrics, subnets = self.load_object_sets(origin)
        table = tables.FabricsTable(visible_columns=visible_columns, subnets=subnets)
        return table.render(target, fabrics)


class cmd_fabric(OriginPagedTableCommand):
    """Details of a fabric."""

    def __init__(self, parser):
        super(cmd_fabric, self).__init__(parser)
        parser.add_argument("name", nargs=1, help=("Name of the fabric."))

    @asynchronous
    async def load_object_sets(self, origin):
        fabrics = origin.Fabrics.read()
        subnets = origin.Subnets.read()
        return await fabrics, await subnets

    def execute(self, origin, options, target):
        try:
            fabric = origin.Fabric.read(options.name[0])
        except CallError as error:
            if error.status == HTTPStatus.NOT_FOUND:
                raise CommandError("Unable to find fabric %s." % options.name[0])
            else:
                raise
        fabrics, subnets = self.load_object_sets(origin)
        table = tables.FabricDetail(fabrics=fabrics, subnets=subnets)
        return table.render(target, fabric)


def register(parser):
    """Register commands with the given parser."""
    cmd_fabrics.register(parser)
    cmd_fabric.register(parser)



"""Test for `maas.client.viscera.devices`."""

from testtools.matchers import Equals, IsInstance

from .. import devices
from ...testing import make_name_without_spaces, TestCase
from ..testing import bind


def make_origin():
    # Create a new origin with Devices and Device. The former refers to the
    # latter via the origin, hence why it must be bound.
    return bind(devices.Devices, devices.Device)


class TestDevice(TestCase):
    def test__string_representation_includes_only_system_id_and_hostname(self):
        device = devices.Device(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
            }
        )
        self.assertThat(
            repr(device),
            Equals(
                "<Device hostname=%(hostname)r system_id=%(system_id)r>" % device._data
            ),
        )

    def test__read(self):
        data = {
            "system_id": make_name_without_spaces("system-id"),
            "hostname": make_name_without_spaces("hostname"),
        }

        origin = make_origin()
        origin.Device._handler.read.return_value = data

        device_observed = origin.Device.read(data["system_id"])
        device_expected = origin.Device(data)
        self.assertThat(device_observed, Equals(device_expected))

    def test__get_power_parameters(self):
        device = make_origin().Device(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
            }
        )
        power_parameters = {"key": make_name_without_spaces("value")}
        device._handler.power_parameters.return_value = power_parameters
        self.assertThat(device.get_power_parameters(), Equals(power_parameters))
        device._handler.power_parameters.assert_called_once_with(
            system_id=device.system_id
        )

    def test__set_power(self):
        orig_power_type = make_name_without_spaces("power_type")
        new_power_type = make_name_without_spaces("power_type")
        device = make_origin().Device(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
                "power_type": orig_power_type,
            }
        )
        power_parameters = {"key": make_name_without_spaces("value")}
        device._handler.update.return_value = {"power_type": new_power_type}
        device.set_power(new_power_type, power_parameters)
        device._handler.update.assert_called_once_with(
            system_id=device.system_id,
            power_type=new_power_type,
            power_parameters=power_parameters,
        )
        self.assertThat(device.power_type, Equals(new_power_type))


class TestDevices(TestCase):
    def test__create(self):
        origin = make_origin()
        Devices = origin.Devices
        Devices._handler.create.return_value = {}
        observed = Devices.create(
            ["00:11:22:33:44:55", "00:11:22:33:44:AA"],
            hostname="new-machine",
            domain="maas",
            zone="zone1",
        )
        self.assertThat(observed, IsInstance(devices.Device))
        Devices._handler.create.assert_called_once_with(
            mac_addresses=["00:11:22:33:44:55", "00:11:22:33:44:AA"],
            hostname="new-machine",
            domain="maas",
            zone="zone1",
        )

    def test__create_no_optional(self):
        origin = make_origin()
        Devices = origin.Devices
        Devices._handler.create.return_value = {}
        observed = Devices.create(["00:11:22:33:44:55", "00:11:22:33:44:AA"])
        self.assertThat(observed, IsInstance(devices.Device))
        Devices._handler.create.assert_called_once_with(
            mac_addresses=["00:11:22:33:44:55", "00:11:22:33:44:AA"]
        )

    def test__read(self):
        data = {
            "system_id": make_name_without_spaces("system-id"),
            "hostname": make_name_without_spaces("hostname"),
        }

        origin = make_origin()
        origin.Devices._handler.read.return_value = [data]

        devices_observed = origin.Devices.read()
        devices_expected = origin.Devices([origin.Device(data)])
        self.assertThat(devices_observed, Equals(devices_expected))



"""Tests for `maas.client.viscera.resource_pools`."""

import random

from testtools.matchers import Equals, IsInstance, MatchesStructure

from .. import resource_pools

from ..testing import bind
from ...testing import make_string_without_spaces, TestCase


def make_origin():
    """
    Create a new origin with ResourcePool and ResourcePools. The former
    refers to the latter via the origin, hence why it must be bound.
    """
    return bind(resource_pools.ResourcePools, resource_pools.ResourcePool)


class TestResourcePools(TestCase):
    def test_resource_pools_create(self):
        origin = make_origin()
        pool_id = random.randint(0, 100)
        name = make_string_without_spaces()
        description = make_string_without_spaces()
        origin.ResourcePools._handler.create.return_value = {
            "id": pool_id,
            "name": name,
            "description": description,
        }
        pool = origin.ResourcePools.create(name=name, description=description)
        origin.ResourcePools._handler.create.assert_called_once_with(
            name=name, description=description
        )
        self.assertThat(pool, IsInstance(origin.ResourcePool))
        self.assertThat(
            pool,
            MatchesStructure.byEquality(id=pool_id, name=name, description=description),
        )

    def test_resource_pools_create_without_description(self):
        origin = make_origin()
        pool_id = random.randint(0, 100)
        name = make_string_without_spaces()
        description = ""
        origin.ResourcePools._handler.create.return_value = {
            "id": pool_id,
            "name": name,
            "description": description,
        }
        pool = origin.ResourcePools.create(name=name, description=description)
        origin.ResourcePools._handler.create.assert_called_once_with(
            name=name, description=description
        )
        self.assertThat(pool, IsInstance(origin.ResourcePool))
        self.assertThat(
            pool,
            MatchesStructure.byEquality(id=pool_id, name=name, description=description),
        )

    def test_resource_pools_read(self):
        ResourcePools = make_origin().ResourcePools
        pools = [
            {
                "id": random.randint(0, 100),
                "name": make_string_without_spaces(),
                "description": make_string_without_spaces(),
            }
            for _ in range(3)
        ]
        ResourcePools._handler.read.return_value = pools
        pools = ResourcePools.read()
        self.assertThat(len(pools), Equals(3))


class TestResourcePool(TestCase):
    def test_resource_pool_read(self):
        ResourcePool = make_origin().ResourcePool
        pool = {
            "id": random.randint(0, 100),
            "name": make_string_without_spaces(),
            "description": make_string_without_spaces(),
        }
        ResourcePool._handler.read.return_value = pool
        self.assertThat(ResourcePool.read(id=pool["id"]), Equals(ResourcePool(pool)))
        ResourcePool._handler.read.assert_called_once_with(id=pool["id"])

    def test_resource_pool_delete(self):
        ResourcePool = make_origin().ResourcePool
        pool_id = random.randint(0, 100)
        pool = ResourcePool(
            {
                "id": pool_id,
                "name": make_string_without_spaces(),
                "description": make_string_without_spaces(),
            }
        )
        pool.delete()
        ResourcePool._handler.delete.assert_called_once_with(id=pool_id)



# Copyright 2016-2017 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Miscellaneous types."""

__all__ = ["JSONArray", "JSONObject", "JSONValue"]

from typing import Dict, Sequence, Union

#
# Types that can be represented in JSON.
#

JSONValue = Union[str, int, float, bool, None, "JSONArray", "JSONObject"]
JSONArray = Sequence[JSONValue]
JSONObject = Dict[str, JSONValue]



"""Tests for `maas.client.bones`."""

import json
import random
from unittest.mock import ANY, Mock
from urllib.parse import parse_qsl, urlparse
from uuid import uuid1

from testtools.matchers import Equals, Is, MatchesStructure

from .. import testing
from ... import bones
from ...testing import TestCase
from ...utils.tests.test_auth import make_Credentials


class TestSessionAPI(TestCase):
    def test__fromURL_raises_SessionError_when_request_fails(self):
        fixture = self.useFixture(testing.DescriptionServer(b"bogus"))
        error = self.assertRaises(
            bones.SessionError, bones.SessionAPI.fromURL, fixture.url + "bogus/"
        )
        self.assertEqual(fixture.url + "bogus/ -> 404 Not Found", str(error))

    def test__fromURL_raises_SessionError_when_content_not_json(self):
        fixture = self.useFixture(testing.DescriptionServer())
        fixture.handler.content_type = "text/json"
        error = self.assertRaises(
            bones.SessionError, bones.SessionAPI.fromURL, fixture.url
        )
        self.assertEqual("Expected application/json, got: text/json", str(error))

    async def test__fromURL_sets_credentials_on_session(self):
        fixture = self.useFixture(testing.DescriptionServer())
        credentials = make_Credentials()
        session = await bones.SessionAPI.fromURL(fixture.url, credentials=credentials)
        self.assertIs(credentials, session.credentials)

    async def test__fromURL_sets_insecure_on_session(self):
        insecure = random.choice((True, False))
        fixture = self.useFixture(testing.DescriptionServer())
        session = await bones.SessionAPI.fromURL(fixture.url, insecure=insecure)
        self.assertThat(session.insecure, Is(insecure))

    async def test__fromURL_sets_scheme_on_session(self):
        insecure = random.choice((True, False))
        fixture = self.useFixture(testing.DescriptionServer())
        session = await bones.SessionAPI.fromURL(fixture.url, insecure=insecure)
        self.assertThat(session.scheme, Equals("http"))


class TestSessionAPI_APIVersions(TestCase):
    """Tests for `SessionAPI` with multiple API versions."""

    scenarios = tuple(
        (name, dict(version=version, path=path))
        for name, version, path in testing.list_api_descriptions()
    )

    async def test__fromURL_downloads_description(self):
        description = self.path.read_bytes()
        fixture = self.useFixture(testing.DescriptionServer(description))
        session = await bones.SessionAPI.fromURL(fixture.url)
        self.assertEqual(json.loads(description.decode("utf-8")), session.description)


class TestActionAPI_APIVersions(TestCase):
    """Tests for `ActionAPI` with multiple API versions."""

    scenarios = tuple(
        (name, dict(version=version, description=description))
        for name, version, description in testing.api_descriptions
    )

    def test__Version_read(self):
        session = bones.SessionAPI(self.description)
        action = session.Version.read
        self.assertThat(
            action,
            MatchesStructure.byEquality(
                name="read",
                fullname="Version.read",
                method="GET",
                handler=session.Version,
                is_restful=True,
                op=None,
            ),
        )

    def test__Machines_deployment_status(self):
        if self.version > (2, 0):
            self.skipTest("Machines.deployment_status only in <= 2.0")

        session = bones.SessionAPI(self.description, ("a", "b", "c"))
        action = session.Machines.deployment_status
        self.assertThat(
            action,
            MatchesStructure.byEquality(
                name="deployment_status",
                fullname="Machines.deployment_status",
                method="GET",
                handler=session.Machines,
                is_restful=False,
                op="deployment_status",
            ),
        )

    def test__Machines_power_parameters(self):
        session = bones.SessionAPI(self.description, ("a", "b", "c"))
        action = session.Machines.power_parameters
        self.assertThat(
            action,
            MatchesStructure.byEquality(
                name="power_parameters",
                fullname="Machines.power_parameters",
                method="GET",
                handler=session.Machines,
                is_restful=False,
                op="power_parameters",
            ),
        )


class TestCallAPI_APIVersions(TestCase):
    """Tests for `CallAPI` with multiple API versions."""

    scenarios = tuple(
        (name, dict(version=version, description=description))
        for name, version, description in testing.api_descriptions
    )

    def test__marshals_lists_into_query_as_repeat_parameters(self):
        system_ids = list(str(uuid1()) for _ in range(3))
        session = bones.SessionAPI(self.description, ("a", "b", "c"))
        call = session.Machines.power_parameters.bind()
        call.dispatch = Mock()

        call.call(nodes=system_ids)

        call.dispatch.assert_called_once_with(ANY, ANY, ANY)
        uri, body, headers = call.dispatch.call_args[0]
        uri = urlparse(uri)
        self.assertThat(uri.path, Equals("/MAAS/api/2.0/machines/"))
        query_expected = [("op", "power_parameters")]
        query_expected.extend(("nodes", system_id) for system_id in system_ids)
        query_observed = parse_qsl(uri.query)
        self.assertThat(query_observed, Equals(query_expected))



"""Objects for machines."""

__all__ = ["Machine", "Machines"]

import asyncio
import base64
import bson
import json
from http import HTTPStatus
import typing

from . import (
    check,
    check_optional,
    ObjectField,
    ObjectFieldRelated,
    ObjectFieldRelatedSet,
    to,
)
from .fabrics import Fabric
from .interfaces import Interface
from .nodes import Node, Nodes, NodesType, NodeTypeMeta
from .pods import Pod
from .subnets import Subnet
from .zones import Zone
from ..bones import CallError
from ..enum import NodeStatus, PowerState, PowerStopMode
from ..errors import MAASException, OperationNotAllowed, PowerError
from ..utils import remove_None
from ..utils.diff import calculate_dict_diff


FabricParam = typing.Union[str, int, Fabric]
InterfaceParam = typing.Union[str, int, Interface]
SubnetParam = typing.Union[str, int, Subnet]
ZoneParam = typing.Union[str, Zone]


def get_param_arg(param, idx, klass, arg, attr="id"):
    """Return the correct value for a fabric from `arg`."""
    if isinstance(arg, klass):
        return getattr(arg, attr)
    elif isinstance(arg, (int, str)):
        return arg
    else:
        raise TypeError(
            "%s[%d] must be int, str, or %s, not %s"
            % (param, idx, klass.__name__, type(arg).__name__)
        )


class MachinesType(NodesType):
    """Metaclass for `Machines`."""

    async def create(
        cls,
        architecture: str,
        mac_addresses: typing.Sequence[str],
        power_type: str,
        power_parameters: typing.Mapping[str, typing.Any] = None,
        *,
        subarchitecture: str = None,
        min_hwe_kernel: str = None,
        hostname: str = None,
        domain: typing.Union[int, str] = None
    ):
        """
        Create a new machine.

        :param architecture: The architecture of the machine (required).
        :type architecture: `str`
        :param mac_addresses: The MAC address of the machine (required).
        :type mac_addresses: sequence of `str`
        :param power_type: The power type of the machine (required).
        :type power_type: `str`
        :param power_parameters: The power parameters for the machine
            (optional).
        :type power_parameters: mapping of `str` to any value.
        :param subarchitecture: The subarchitecture of the machine (optional).
        :type subarchitecture: `str`
        :param min_hwe_kernel: The minimal HWE kernel for the machine
            (optional).
        :param hostname: The hostname for the machine (optional).
        :type hostname: `str`
        :param domain: The domain for the machine (optional).
        :type domain: `int` or `str`
        """
        params = {
            "architecture": architecture,
            "mac_addresses": mac_addresses,
            "power_type": power_type,
        }
        if power_parameters is not None:
            params["power_parameters"] = json.dumps(power_parameters, sort_keys=True)
        if subarchitecture is not None:
            params["subarchitecture"] = subarchitecture
        if min_hwe_kernel is not None:
            params["min_hwe_kernel"] = min_hwe_kernel
        if hostname is not None:
            params["hostname"] = hostname
        if domain is not None:
            params["domain"] = domain
        return cls._object(await cls._handler.create(**params))

    async def allocate(
        cls,
        *,
        hostname: str = None,
        architectures: typing.Sequence[str] = None,
        cpus: int = None,
        fabrics: typing.Sequence[FabricParam] = None,
        interfaces: typing.Sequence[InterfaceParam] = None,
        memory: float = None,
        pod: typing.Union[str, Pod] = None,
        not_pod: typing.Union[str, Pod] = None,
        pod_type: str = None,
        not_pod_type: str = None,
        storage: typing.Sequence[str] = None,
        subnets: typing.Sequence[SubnetParam] = None,
        tags: typing.Sequence[str] = None,
        zone: typing.Union[str, Zone] = None,
        not_fabrics: typing.Sequence[FabricParam] = None,
        not_subnets: typing.Sequence[SubnetParam] = None,
        not_tags: typing.Sequence[str] = None,
        not_zones: typing.Sequence[ZoneParam] = None,
        agent_name: str = None,
        comment: str = None,
        bridge_all: bool = None,
        bridge_stp: bool = None,
        bridge_fd: int = None,
        dry_run: bool = None,
        verbose: bool = None
    ):
        """
        Allocate a machine.

        :param hostname: The hostname to match.
        :type hostname: `str`
        :param architectures: The architecture(s) to match.
        :type architectures: sequence of `str`
        :param cpus: The minimum number of CPUs to match.
        :type cpus: `int`
        :param fabrics: The connected fabrics to match.
        :type fabrics: sequence of either `str`, `int`, or `Fabric`
        :param interfaces: The interfaces to match.
        :type interfaces: sequence of either `str`, `int`, or `Interface`
        :param memory: The minimum amount of RAM to match in MiB.
        :type memory: `int`
        :param pod: The pod to allocate the machine from.
        :type pod: `str`
        :param not_pod: Pod the machine must not be located in.
        :type not_pod: `str`
        :param pod_type: The type of pod to allocate the machine from.
        :type pod_type: `str`
        :param not_pod_type: Pod type the machine must not be located in.
        :type not_pod_type: `str`
        :param subnets: The subnet(s) the desired machine must be linked to.
        :type subnets: sequence of `str` or `int` or `Subnet`
        :param storage: The storage contraint to match.
        :type storage: `str`
        :param tags: The tags to match, as a sequence.
        :type tags: sequence of `str`
        :param zone: The zone the desired machine must belong to.
        :type zone: `str` or `Zone`
        :param not_fabrics: The fabrics the machine must NOT be connected to.
        :type not_fabrics: sequence of either `str`, `int`, or `Fabric`
        :param not_subnets: The subnet(s) the desired machine must NOT be
            linked to.
        :type not_subnets: sequence of `str` or `int` or `Subnet`
        :param not_zones: The zone(s) the desired machine must NOT in.
        :type not_zones: sequence of `str` or `Zone`
        :param agent_name: Agent name to attach to the acquire machine.
        :type agent_name: `str`
        :param comment: Comment for the allocate event placed on machine.
        :type comment: `str`
        :param bridge_all: Automatically create a bridge on all interfaces
            on the allocated machine.
        :type bridge_all: `bool`
        :param bridge_stp: Turn spaning tree protocol on or off for the
            bridges created with bridge_all.
        :type bridge_stp: `bool`
        :param bridge_fd: Set the forward delay in seconds on the bridges
            created with bridge_all.
        :type bridge_fd: `int`
        :param dry_run: Don't actually acquire the machine just return the
            machine that would have been acquired.
        :type dry_run: `bool`
        :param verbose: Indicate that the user would like additional verbosity
            in the constraints_by_type field (each constraint will be prefixed
            by `verbose_`, and contain the full data structure that indicates
            which machine(s) matched).
        :type verbose: `bool`
        """
        params = remove_None(
            {
                "name": hostname,
                "arch": architectures,
                "cpu_count": str(cpus) if cpus else None,
                "mem": str(memory) if memory else None,
                "pod_type": pod_type,
                "not_pod_type": not_pod_type,
                "storage": storage,
                "tags": tags,
                "not_tags": not_tags,
                "agent_name": agent_name,
                "comment": comment,
                "bridge_all": bridge_all,
                "bridge_stp": bridge_stp,
                "bridge_fd": bridge_fd,
                "dry_run": dry_run,
                "verbose": verbose,
            }
        )
        if fabrics is not None:
            params["fabrics"] = [
                get_param_arg("fabrics", idx, Fabric, fabric)
                for idx, fabric in enumerate(fabrics)
            ]
        if interfaces is not None:
            params["interfaces"] = [
                get_param_arg("interfaces", idx, Interface, nic)
                for idx, nic in enumerate(interfaces)
            ]
        if pod is not None:
            if isinstance(pod, Pod):
                params["pod"] = pod.name
            elif isinstance(pod, str):
                params["pod"] = pod
            else:
                raise TypeError("pod must be a str or Pod, not %s" % type(pod).__name__)
        if not_pod is not None:
            if isinstance(not_pod, Pod):
                params["not_pod"] = not_pod.name
            elif isinstance(not_pod, str):
                params["not_pod"] = not_pod
            else:
                raise TypeError(
                    "not_pod must be a str or Pod, not %s" % type(not_pod).__name__
                )
        if subnets is not None:
            params["subnets"] = [
                get_param_arg("subnets", idx, Subnet, subnet)
                for idx, subnet in enumerate(subnets)
            ]
        if zone is not None:
            if isinstance(zone, Zone):
                params["zone"] = zone.name
            elif isinstance(zone, str):
                params["zone"] = zone
            else:
                raise TypeError(
                    "zone must be a str or Zone, not %s" % type(zone).__name__
                )
        if not_fabrics is not None:
            params["not_fabrics"] = [
                get_param_arg("not_fabrics", idx, Fabric, fabric)
                for idx, fabric in enumerate(not_fabrics)
            ]
        if not_subnets is not None:
            params["not_subnets"] = [
                get_param_arg("not_subnets", idx, Subnet, subnet)
                for idx, subnet in enumerate(not_subnets)
            ]
        if not_zones is not None:
            params["not_in_zones"] = [
                get_param_arg("not_zones", idx, Zone, zone, attr="name")
                for idx, zone in enumerate(not_zones)
            ]
        try:
            data = await cls._handler.allocate(**params)
        except CallError as error:
            if error.status == HTTPStatus.CONFLICT:
                message = "No machine matching the given criteria was found."
                raise MachineNotFound(message) from error
            else:
                raise
        else:
            return cls._object(data)

    async def get_power_parameters_for(cls, system_ids: typing.Sequence[str]):
        """
        Get a list of power parameters for specified systems.
        *WARNING*: This method is considered 'alpha' and may be modified
        in future.

        :param system_ids: The system IDs to get power parameters for
        """
        if len(system_ids) == 0:
            return {}
        data = await cls._handler.power_parameters(id=system_ids)
        return data


class MachineNotFound(Exception):
    """
    Machine was not found.

    Not a MAASException because this doesn't occur in the context of
    a specific object.
    """


class RescueModeFailure(MAASException):
    """Machine failed to perform a Rescue mode transition."""


class FailedCommissioning(MAASException):
    """Machine failed to commission."""


class FailedTesting(MAASException):
    """Machine failed testing."""


class FailedDeployment(MAASException):
    """Machine failed to deploy."""


class FailedReleasing(MAASException):
    """Machine failed to release."""


class FailedDiskErasing(MAASException):
    """Machine failed to erase disk when releasing."""


class Machines(Nodes, metaclass=MachinesType):
    """The set of machines stored in MAAS."""


class MachineType(NodeTypeMeta):
    """Metaclass for `Machine`."""

    async def read(cls, system_id):
        data = await cls._handler.read(system_id=system_id)
        return cls(data)


class Machine(Node, metaclass=MachineType):
    """A machine stored in MAAS."""

    architecture = ObjectField.Checked(
        "architecture", check_optional(str), check_optional(str)
    )
    boot_disk = ObjectFieldRelated("boot_disk", "BlockDevice", readonly=True)
    boot_interface = ObjectFieldRelated("boot_interface", "Interface", readonly=True)
    block_devices = ObjectFieldRelatedSet(
        "blockdevice_set", "BlockDevices", reverse=None
    )
    bcaches = ObjectFieldRelatedSet("bcaches", "Bcaches", reverse=None)
    cache_sets = ObjectFieldRelatedSet("cache_sets", "BcacheCacheSets", reverse=None)
    cpus = ObjectField.Checked("cpu_count", check(int), check(int))
    disable_ipv4 = ObjectField.Checked("disable_ipv4", check(bool), check(bool))
    distro_series = ObjectField.Checked("distro_series", check(str), readonly=True)
    hwe_kernel = ObjectField.Checked(
        "hwe_kernel", check_optional(str), check_optional(str)
    )
    locked = ObjectField.Checked("locked", check(bool), readonly=True)
    memory = ObjectField.Checked("memory", check(int), check(int))
    min_hwe_kernel = ObjectField.Checked(
        "min_hwe_kernel", check_optional(str), check_optional(str)
    )
    netboot = ObjectField.Checked("netboot", check(bool), readonly=True)
    osystem = ObjectField.Checked("osystem", check(str), readonly=True)
    owner_data = ObjectField.Checked("owner_data", check(dict), check(dict))
    status = ObjectField.Checked("status", to(NodeStatus), readonly=True)
    status_action = ObjectField.Checked(
        "status_action", check_optional(str), readonly=True
    )
    status_message = ObjectField.Checked(
        "status_message", check_optional(str), readonly=True
    )
    status_name = ObjectField.Checked("status_name", check(str), readonly=True)
    raids = ObjectFieldRelatedSet("raids", "Raids", reverse=None)
    volume_groups = ObjectFieldRelatedSet("volume_groups", "VolumeGroups", reverse=None)

    async def save(self):
        """Save the machine in MAAS."""
        orig_owner_data = self._orig_data["owner_data"]
        new_owner_data = dict(self._data["owner_data"])
        self._changed_data.pop("owner_data", None)
        await super(Machine, self).save()
        params_diff = calculate_dict_diff(orig_owner_data, new_owner_data)
        if len(params_diff) > 0:
            params_diff["system_id"] = self.system_id
            await self._handler.set_owner_data(**params_diff)
            self._data["owner_data"] = self._data["owner_data"]

    async def abort(self, *, comment: str = None):
        """Abort the current action.

        :param comment: Reason for aborting the action.
        :param type: `str`
        """
        params = {"system_id": self.system_id}
        if comment:
            params["comment"] = comment
        self._reset(await self._handler.abort(**params))
        return self

    async def clear_default_gateways(self):
        """Clear default gateways."""
        self._reset(
            await self._handler.clear_default_gateways(system_id=self.system_id)
        )
        return self

    async def commission(
        self,
        *,
        enable_ssh: bool = None,
        skip_networking: bool = None,
        skip_storage: bool = None,
        commissioning_scripts: typing.Sequence[str] = None,
        testing_scripts: typing.Sequence[str] = None,
        wait: bool = False,
        wait_interval: int = 5
    ):
        """Commission this machine.

        :param enable_ssh: Prevent the machine from powering off after running
            commissioning scripts and enable your user to SSH into the machine.
        :type enable_ssh: `bool`
        :param skip_networking: Skip updating the MAAS interfaces for the
            machine.
        :type skip_networking: `bool`
        :param skip_storage: Skip update the MAAS block devices for the
            machine.
        :type skip_storage: `bool`
        :param commissioning_scripts: List of extra commisisoning scripts
            to run. If the name of the commissioning scripts match a tag, then
            all commissioning scripts with that tag will be used.
        :type commissioning_scripts: sequence of `str`
        :param testing_scripts: List of testing scripts to run after
            commissioning. By default a small set of testing scripts will run
            by default. Passing empty list will disable running any testing
            scripts during commissioning. If the name of the testing scripts
            match a tag, then all testing scripts with that tag will be used.
        :type testing_scripts: sequence of `str`
        :param wait: If specified, wait until the commissioning is complete.
        :param wait_interval: How often to poll, defaults to 5 seconds
        """
        params = {"system_id": self.system_id}
        if enable_ssh is not None:
            params["enable_ssh"] = enable_ssh
        if skip_networking is not None:
            params["skip_networking"] = skip_networking
        if skip_storage is not None:
            params["skip_storage"] = skip_storage
        if commissioning_scripts is not None and len(commissioning_scripts) > 0:
            params["commissioning_scripts"] = ",".join(commissioning_scripts)
        if testing_scripts is not None:
            if len(testing_scripts) == 0 or testing_scripts == "none":
                params["testing_scripts"] = ["none"]
            else:
                params["testing_scripts"] = ",".join(testing_scripts)
        self._reset(await self._handler.commission(**params))
        if not wait:
            return self
        else:
            # Wait for the machine to be fully commissioned.
            while self.status in [NodeStatus.COMMISSIONING, NodeStatus.TESTING]:
                await asyncio.sleep(wait_interval)
                self._reset(await self._handler.read(system_id=self.system_id))
            if self.status == NodeStatus.FAILED_COMMISSIONING:
                msg = "{hostname} failed to commission.".format(hostname=self.hostname)
                raise FailedCommissioning(msg, self)
            if self.status == NodeStatus.FAILED_TESTING:
                msg = "{hostname} failed testing.".format(hostname=self.hostname)
                raise FailedTesting(msg, self)
            return self

    async def deploy(
        self,
        *,
        user_data: typing.Union[bytes, str] = None,
        distro_series: str = None,
        hwe_kernel: str = None,
        comment: str = None,
        wait: bool = False,
        install_kvm: bool = False,
        wait_interval: int = 5
    ):
        """Deploy this machine.

        :param user_data: User-data to provide to the machine when booting. If
            provided as a byte string, it will be base-64 encoded prior to
            transmission. If provided as a Unicode string it will be assumed
            to be already base-64 encoded.
        :param distro_series: The OS to deploy.
        :param hwe_kernel: The HWE kernel to deploy. Probably only relevant
            when deploying Ubuntu.
        :param comment: A comment for the event log.
        :param wait: If specified, wait until the deploy is complete.
        :param wait_interval: How often to poll, defaults to 5 seconds
        """
        params = {"system_id": self.system_id}

        if install_kvm:
            params["install_kvm"] = install_kvm

        if user_data is not None:
            if isinstance(user_data, bytes):
                params["user_data"] = base64.encodebytes(user_data)
            else:
                # Already base-64 encoded. Convert to a byte string in
                # preparation for multipart assembly.
                params["user_data"] = user_data.encode("ascii")
        if distro_series is not None:
            params["distro_series"] = distro_series
        if hwe_kernel is not None:
            params["hwe_kernel"] = hwe_kernel
        if comment is not None:
            params["comment"] = comment
        self._reset(await self._handler.deploy(**params))
        if not wait:
            return self
        else:
            # Wait for the machine to be fully deployed
            while self.status == NodeStatus.DEPLOYING:
                await asyncio.sleep(wait_interval)
                self._reset(await self._handler.read(system_id=self.system_id))
            if self.status == NodeStatus.FAILED_DEPLOYMENT:
                msg = "{hostname} failed to deploy.".format(hostname=self.hostname)
                raise FailedDeployment(msg, self)
            return self

    async def enter_rescue_mode(self, wait: bool = False, wait_interval: int = 5):
        """
        Send this machine into 'rescue mode'.

        :param wait: If specified, wait until the deploy is complete.
        :param wait_interval: How often to poll, defaults to 5 seconds
        """
        try:
            self._reset(await self._handler.rescue_mode(system_id=self.system_id))
        except CallError as error:
            if error.status == HTTPStatus.FORBIDDEN:
                message = "Not allowed to enter rescue mode"
                raise OperationNotAllowed(message) from error
            else:
                raise

        if not wait:
            return self
        else:
            # Wait for machine to finish entering rescue mode
            while self.status == NodeStatus.ENTERING_RESCUE_MODE:
                await asyncio.sleep(wait)
                self._reset(await self._handler.read(system_id=self.system_id))
            if self.status == NodeStatus.FAILED_ENTERING_RESCUE_MODE:
                msg = "{hostname} failed to enter rescue mode.".format(
                    hostname=self.hostname
                )
                raise RescueModeFailure(msg, self)
            return self

    async def exit_rescue_mode(self, wait: bool = False, wait_interval: int = 5):
        """
        Exit rescue mode.

        :param wait: If specified, wait until the deploy is complete.
        :param wait_interval: How often to poll, defaults to 5 seconds
        """
        try:
            self._reset(await self._handler.exit_rescue_mode(system_id=self.system_id))
        except CallError as error:
            if error.status == HTTPStatus.FORBIDDEN:
                message = "Not allowed to exit rescue mode."
                raise OperationNotAllowed(message) from error
            else:
                raise
        if not wait:
            return self
        else:
            # Wait for machine to finish exiting rescue mode
            while self.status == NodeStatus.EXITING_RESCUE_MODE:
                await asyncio.sleep(wait_interval)
                self._reset(await self._handler.read(system_id=self.system_id))
            if self.status == NodeStatus.FAILED_EXITING_RESCUE_MODE:
                msg = "{hostname} failed to exit rescue mode.".format(
                    hostname=self.hostname
                )
                raise RescueModeFailure(msg, self)
            return self

    async def get_curtin_config(self):
        """Get the curtin configuration.

        :returns: Curtin configuration
        :rtype: `str`
        """
        return self._handler.get_curtin_config(system_id=self.system_id)

    async def get_details(self):
        """Get machine details information.

        :returns: Mapping of hardware details.
        """
        data = await self._handler.details(system_id=self.system_id)
        return bson.decode_all(data)[0]

    async def mark_broken(self, *, comment: str = None):
        """Mark broken.

        :param comment: Reason machine is broken.
        :type comment: `str`
        """
        params = {"system_id": self.system_id}
        if comment:
            params["comment"] = comment
        self._reset(await self._handler.mark_broken(**params))
        return self

    async def mark_fixed(self, *, comment: str = None):
        """Mark fixes.

        :param comment: Reason machine is fixed.
        :type comment: `str`
        """
        params = {"system_id": self.system_id}
        if comment:
            params["comment"] = comment
        self._reset(await self._handler.mark_fixed(**params))
        return self

    async def release(
        self,
        *,
        comment: str = None,
        erase: bool = None,
        secure_erase: bool = None,
        quick_erase: bool = None,
        wait: bool = False,
        wait_interval: int = 5
    ):
        """
        Release the machine.

        :param comment: Reason machine was released.
        :type comment: `str`
        :param erase: Erase the disk when release.
        :type erase: `bool`
        :param secure_erase: Use the drive's secure erase feature if available.
        :type secure_erase: `bool`
        :param quick_erase: Wipe the just the beginning and end of the disk.
            This is not secure.
        :param wait: If specified, wait until the deploy is complete.
        :type wait: `bool`
        :param wait_interval: How often to poll, defaults to 5 seconds.
        :type wait_interval: `int`
        """
        params = remove_None(
            {
                "system_id": self.system_id,
                "comment": comment,
                "erase": erase,
                "secure_erase": secure_erase,
                "quick_erase": quick_erase,
            }
        )
        self._reset(await self._handler.release(**params))
        if not wait:
            return self
        else:
            # Wait for machine to be released
            while self.status in [NodeStatus.RELEASING, NodeStatus.DISK_ERASING]:
                await asyncio.sleep(wait_interval)
                try:
                    self._reset(await self._handler.read(system_id=self.system_id))
                except CallError as error:
                    if error.status == HTTPStatus.NOT_FOUND:
                        # Release must have been on a machine in a pod. This
                        # machine no longer exists. Just return the machine
                        # as it has been released.
                        return self
                    else:
                        raise
            if self.status == NodeStatus.FAILED_RELEASING:
                msg = "{hostname} failed to be released.".format(hostname=self.hostname)
                raise FailedReleasing(msg, self)
            elif self.status == NodeStatus.FAILED_DISK_ERASING:
                msg = "{hostname} failed to erase disk.".format(hostname=self.hostname)
                raise FailedDiskErasing(msg, self)
            return self

    async def power_on(
        self, comment: str = None, wait: bool = False, wait_interval: int = 5
    ):
        """
        Power on.

        :param comment: Reason machine was powered on.
        :type comment: `str`
        :param wait: If specified, wait until the machine is powered on.
        :type wait: `bool`
        :param wait_interval: How often to poll, defaults to 5 seconds.
        :type wait_interval: `int`
        """
        params = {"system_id": self.system_id}
        if comment is not None:
            params["comment"] = comment
        try:
            self._reset(await self._handler.power_on(**params))
        except CallError as error:
            if error.status == HTTPStatus.FORBIDDEN:
                message = "Not allowed to power on machine."
                raise OperationNotAllowed(message) from error
            else:
                raise
        if not wait or self.power_state == PowerState.UNKNOWN:
            # Don't wait for a machine that always shows power state as
            # unknown as the driver cannot query the power state.
            return self
        else:
            # Wait for machine to be powered on.
            while self.power_state == PowerState.OFF:
                await asyncio.sleep(wait_interval)
                self._reset(await self._handler.read(system_id=self.system_id))
            if self.power_state == PowerState.ERROR:
                msg = "{hostname} failed to power on.".format(hostname=self.hostname)
                raise PowerError(msg, self)
            return self

    async def power_off(
        self,
        stop_mode: PowerStopMode = PowerStopMode.HARD,
        comment: str = None,
        wait: bool = False,
        wait_interval: int = 5,
    ):
        """
        Power off.

        :param stop_mode: How to perform the power off.
        :type stop_mode: `PowerStopMode`
        :param comment: Reason machine was powered on.
        :type comment: `str`
        :param wait: If specified, wait until the machine is powered on.
        :type wait: `bool`
        :param wait_interval: How often to poll, defaults to 5 seconds.
        :type wait_interval: `int`
        """
        params = {"system_id": self.system_id, "stop_mode": stop_mode.value}
        if comment is not None:
            params["comment"] = comment
        try:
            self._reset(await self._handler.power_off(**params))
        except CallError as error:
            if error.status == HTTPStatus.FORBIDDEN:
                message = "Not allowed to power off machine."
                raise OperationNotAllowed(message) from error
            else:
                raise
        if not wait or self.power_state == PowerState.UNKNOWN:
            # Don't wait for a machine that always shows power state as
            # unknown as the driver cannot query the power state.
            return self
        else:
            # Wait for machine to be powered off.
            while self.power_state == PowerState.ON:
                await asyncio.sleep(wait_interval)
                self._reset(await self._handler.read(system_id=self.system_id))
            if self.power_state == PowerState.ERROR:
                msg = "{hostname} failed to power off.".format(hostname=self.hostname)
                raise PowerError(msg, self)
            return self

    async def query_power_state(self):
        """
        Query the machine's BMC for the current power state.

        :returns: Current power state.
        :rtype: `PowerState`
        """
        power_data = await self._handler.query_power_state(system_id=self.system_id)
        # Update the internal state of this object as well, since we have the
        # updated power state from the BMC directly. MAAS server does this as
        # well, just do it client side to make it nice for a developer.
        self._data["power_state"] = power_data["state"]
        return PowerState(power_data["state"])

    async def restore_default_configuration(self):
        """
        Restore machine's configuration to its initial state.
        """
        self._reset(
            await self._handler.restore_default_configuration(system_id=self.system_id)
        )

    async def restore_networking_configuration(self):
        """
        Restore machine's networking configuration to its initial state.
        """
        self._reset(
            await self._handler.restore_networking_configuration(
                system_id=self.system_id
            )
        )

    async def restore_storage_configuration(self):
        """
        Restore machine's storage configuration to its initial state.
        """
        self._reset(
            await self._handler.restore_storage_configuration(system_id=self.system_id)
        )

    async def lock(self, *, comment: str = None):
        """Lock the machine to prevent changes.

        :param comment: Reason machine was locked.
        :type comment: `str`
        """
        params = {"system_id": self.system_id}
        if comment:
            params["comment"] = comment
        self._reset(await self._handler.lock(**params))
        return self

    async def unlock(self, *, comment: str = None):
        """Unlock the machine allowing changes.

        :param comment: Reason machine was unlocked.
        :type comment: `str`
        """
        params = {"system_id": self.system_id}
        if comment:
            params["comment"] = comment
        self._reset(await self._handler.unlock(**params))
        return self



"""Tests for `maas.client._client`."""

from unittest.mock import Mock

from testtools.matchers import IsInstance, MatchesAll, MatchesStructure

from .. import facade, viscera
from ..testing import TestCase


class TestClient(TestCase):
    """Tests for the simplified client.

    Right now these are fairly trivial tests, not testing in depth. Work is in
    progress to create a unit test framework that will allow testing against
    fake MAAS servers that match the "shape" of real MAAS servers. At the
    point that lands these tests should be revised.
    """

    def setUp(self):
        super(TestClient, self).setUp()
        self.session = Mock(name="session", handlers={})
        self.origin = viscera.Origin(self.session)
        self.client = facade.Client(self.origin)

    def test__client_maps_account(self):
        self.assertThat(
            self.client,
            MatchesClient(
                account=MatchesFacade(
                    create_credentials=self.origin.Account.create_credentials,
                    delete_credentials=self.origin.Account.delete_credentials,
                )
            ),
        )

    def test__client_maps_boot_resources(self):
        self.assertThat(
            self.client,
            MatchesClient(
                boot_resources=MatchesFacade(
                    create=self.origin.BootResources.create,
                    get=self.origin.BootResource.read,
                    list=self.origin.BootResources.read,
                    start_import=self.origin.BootResources.start_import,
                    stop_import=self.origin.BootResources.stop_import,
                )
            ),
        )

    def test__client_maps_boot_sources(self):
        self.assertThat(
            self.client,
            MatchesClient(
                boot_sources=MatchesFacade(
                    create=self.origin.BootSources.create,
                    get=self.origin.BootSource.read,
                    list=self.origin.BootSources.read,
                )
            ),
        )

    def test__client_maps_devices(self):
        self.assertThat(
            self.client,
            MatchesClient(
                devices=MatchesFacade(
                    create=self.origin.Devices.create,
                    get=self.origin.Device.read,
                    list=self.origin.Devices.read,
                )
            ),
        )

    def test__client_maps_events(self):
        self.assertThat(
            self.client,
            MatchesClient(
                events=MatchesFacade(
                    query=self.origin.Events.query,
                    DEBUG=self.origin.Events.Level.DEBUG,
                    INFO=self.origin.Events.Level.INFO,
                    WARNING=self.origin.Events.Level.WARNING,
                    ERROR=self.origin.Events.Level.ERROR,
                    CRITICAL=self.origin.Events.Level.CRITICAL,
                )
            ),
        )

    def test__client_maps_fabrics(self):
        self.assertThat(
            self.client,
            MatchesClient(
                fabrics=MatchesFacade(
                    create=self.origin.Fabrics.create,
                    get=self.origin.Fabric.read,
                    get_default=self.origin.Fabric.get_default,
                    list=self.origin.Fabrics.read,
                )
            ),
        )

    def test__client_maps_subnets(self):
        self.assertThat(
            self.client,
            MatchesClient(
                subnets=MatchesFacade(
                    create=self.origin.Subnets.create,
                    get=self.origin.Subnet.read,
                    list=self.origin.Subnets.read,
                )
            ),
        )

    def test__client_maps_spaces(self):
        self.assertThat(
            self.client,
            MatchesClient(
                spaces=MatchesFacade(
                    create=self.origin.Spaces.create,
                    get=self.origin.Space.read,
                    get_default=self.origin.Space.get_default,
                    list=self.origin.Spaces.read,
                )
            ),
        )

    def test__client_maps_ip_ranges(self):
        self.assertThat(
            self.client,
            MatchesClient(
                ip_ranges=MatchesFacade(
                    create=self.origin.IPRanges.create,
                    get=self.origin.IPRange.read,
                    list=self.origin.IPRanges.read,
                )
            ),
        )

    def test__client_maps_static_routes(self):
        self.assertThat(
            self.client,
            MatchesClient(
                static_routes=MatchesFacade(
                    create=self.origin.StaticRoutes.create,
                    get=self.origin.StaticRoute.read,
                    list=self.origin.StaticRoutes.read,
                )
            ),
        )

    def test__client_maps_files(self):
        self.assertThat(
            self.client, MatchesClient(files=MatchesFacade(list=self.origin.Files.read))
        )

    def test__client_maps_machines(self):
        self.assertThat(
            self.client,
            MatchesClient(
                machines=MatchesFacade(
                    allocate=self.origin.Machines.allocate,
                    create=self.origin.Machines.create,
                    get=self.origin.Machine.read,
                    list=self.origin.Machines.read,
                )
            ),
        )

    def test__client_maps_rack_controllers(self):
        self.assertThat(
            self.client,
            MatchesClient(
                rack_controllers=MatchesFacade(
                    get=self.origin.RackController.read,
                    list=self.origin.RackControllers.read,
                )
            ),
        )

    def test__client_maps_region_controllers(self):
        self.assertThat(
            self.client,
            MatchesClient(
                region_controllers=MatchesFacade(
                    get=self.origin.RegionController.read,
                    list=self.origin.RegionControllers.read,
                )
            ),
        )

    def test__client_maps_ssh_keys(self):
        self.assertThat(
            self.client,
            MatchesClient(
                ssh_keys=MatchesFacade(
                    create=self.origin.SSHKeys.create,
                    get=self.origin.SSHKey.read,
                    list=self.origin.SSHKeys.read,
                )
            ),
        )

    def test__client_maps_tags(self):
        self.assertThat(
            self.client,
            MatchesClient(
                tags=MatchesFacade(
                    create=self.origin.Tags.create, list=self.origin.Tags.read
                )
            ),
        )

    def test__client_maps_users(self):
        self.assertThat(
            self.client,
            MatchesClient(
                users=MatchesFacade(
                    create=self.origin.Users.create,
                    list=self.origin.Users.read,
                    whoami=self.origin.Users.whoami,
                )
            ),
        )

    def test__client_maps_version(self):
        self.assertThat(
            self.client,
            MatchesClient(version=MatchesFacade(get=self.origin.Version.read)),
        )

    def test__client_maps_zones(self):
        self.assertThat(
            self.client,
            MatchesClient(
                zones=MatchesFacade(
                    create=self.origin.Zones.create,
                    get=self.origin.Zone.read,
                    list=self.origin.Zones.read,
                )
            ),
        )

    def test__client_maps_pods(self):
        self.assertThat(
            self.client,
            MatchesClient(
                pods=MatchesFacade(
                    create=self.origin.Pods.create,
                    get=self.origin.Pod.read,
                    list=self.origin.Pods.read,
                )
            ),
        )

    def test__client_maps_resource_pools(self):
        self.assertThat(
            self.client,
            MatchesClient(
                resource_pools=MatchesFacade(
                    create=self.origin.ResourcePools.create,
                    get=self.origin.ResourcePool.read,
                    list=self.origin.ResourcePools.read,
                )
            ),
        )


def MatchesClient(**facades):
    """Matches a `facade.Client` with the given facades."""
    return MatchesAll(
        IsInstance(facade.Client), MatchesStructure(**facades), first_only=True
    )


def MatchesFacade(**methods):
    """Matches a `facade.Facade` with the given methods."""
    return MatchesAll(
        IsInstance(facade.Facade),
        MatchesStructure.byEquality(**methods),
        first_only=True,
    )



"""Objects for nodes."""

__all__ = ["Node", "Nodes"]

try:
    # Python <= 3.9
    from collections import Sequence
except:
    # Python > 3.9
    from collections.abc import Sequence

import typing

from . import (
    check,
    Object,
    ObjectField,
    ObjectFieldRelated,
    ObjectFieldRelatedSet,
    ObjectSet,
    ObjectType,
    to,
)
from ..enum import NodeType, PowerState


def normalize_hostname(hostname):
    """Strips the FQDN from the hostname, since hostname is unique in MAAS."""
    if hostname:
        return hostname.split(".", 1)[0]
    return hostname


def map_tag_name_to_dict(instance, value):
    """Convert a tag name into a dictionary for Tag."""
    return {"name": value, "__incomplete__": True}


class NodesType(ObjectType):
    """Metaclass for `Nodes`."""

    async def read(cls, *, hostnames: typing.Sequence[str] = None):
        """List nodes.

        :param hostnames: Sequence of hostnames to only return.
        :type hostnames: sequence of `str`
        """
        params = {}
        if hostnames:
            params["hostname"] = [
                normalize_hostname(hostname) for hostname in hostnames
            ]
        data = await cls._handler.read(**params)
        return cls(map(cls._object, data))


class Nodes(ObjectSet, metaclass=NodesType):
    """The set of nodes stored in MAAS."""


class NodeTypeMeta(ObjectType):
    """Metaclass for `Node`."""

    async def read(cls, system_id):
        data = await cls._handler.read(system_id=system_id)
        return cls(data)


class Node(Object, metaclass=NodeTypeMeta):
    """A node stored in MAAS."""

    domain = ObjectFieldRelated("domain", "Domain")
    fqdn = ObjectField.Checked("fqdn", check(str), readonly=True)
    hostname = ObjectField.Checked("hostname", check(str), check(str))
    interfaces = ObjectFieldRelatedSet("interface_set", "Interfaces")
    ip_addresses = ObjectField.Checked(  # List[str]
        "ip_addresses", check(Sequence), readonly=True
    )
    node_type = ObjectField.Checked("node_type", to(NodeType), readonly=True)
    owner = ObjectFieldRelated("owner", "User")
    power_state = ObjectField.Checked("power_state", to(PowerState), readonly=True)
    power_type = ObjectField.Checked("power_type", check(str))
    pool = ObjectFieldRelated("pool", "ResourcePool", use_data_setter=True)
    system_id = ObjectField.Checked("system_id", check(str), readonly=True, pk=True)
    tags = ObjectFieldRelatedSet(
        "tag_names", "Tags", reverse=None, map_func=map_tag_name_to_dict
    )
    zone = ObjectFieldRelated("zone", "Zone")

    def __repr__(self):
        return super(Node, self).__repr__(fields={"system_id", "hostname"})

    def as_machine(self):
        """Convert to a `Machine` object.

        `node_type` must be `NodeType.MACHINE`.
        """
        if self.node_type != NodeType.MACHINE:
            raise ValueError("Cannot convert to `Machine`, node_type is not a machine.")
        return self._origin.Machine(self._data)

    def as_device(self):
        """Convert to a `Device` object.

        `node_type` must be `NodeType.DEVICE`.
        """
        if self.node_type != NodeType.DEVICE:
            raise ValueError("Cannot convert to `Device`, node_type is not a device.")
        return self._origin.Device(self._data)

    def as_rack_controller(self):
        """Convert to a `RackController` object.

        `node_type` must be `NodeType.RACK_CONTROLLER` or
        `NodeType.REGION_AND_RACK_CONTROLLER`.
        """
        if self.node_type not in [
            NodeType.RACK_CONTROLLER,
            NodeType.REGION_AND_RACK_CONTROLLER,
        ]:
            raise ValueError(
                "Cannot convert to `RackController`, node_type is not a "
                "rack controller."
            )
        return self._origin.RackController(self._data)

    def as_region_controller(self):
        """Convert to a `RegionController` object.

        `node_type` must be `NodeType.REGION_CONTROLLER` or
        `NodeType.REGION_AND_RACK_CONTROLLER`.
        """
        if self.node_type not in [
            NodeType.REGION_CONTROLLER,
            NodeType.REGION_AND_RACK_CONTROLLER,
        ]:
            raise ValueError(
                "Cannot convert to `RegionController`, node_type is not a "
                "region controller."
            )
        return self._origin.RegionController(self._data)

    async def get_power_parameters(self):
        """Get the power paramters for this node."""
        data = await self._handler.power_parameters(system_id=self.system_id)
        return data

    async def set_power(
        self, power_type: str, power_parameters: typing.Mapping[str, typing.Any] = {}
    ):
        """Set the power type and power parameters for this node."""
        data = await self._handler.update(
            system_id=self.system_id,
            power_type=power_type,
            power_parameters=power_parameters,
        )
        self.power_type = data["power_type"]

    async def save(self):
        # the resource pool uses the name in the API, not the id. The field is
        # defined with use_data_setter=True, so the value in self._changed_data
        # is the full data dict, not just the id.
        if "pool" in self._changed_data:
            self._changed_data["pool"] = self._changed_data["pool"]["name"]
        return await super().save()

    async def delete(self):
        """Deletes the node from MAAS."""
        await self._handler.delete(system_id=self.system_id)



# Copyright 2016-2017 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for handling of MAAS API credentials."""

from testtools.matchers import IsInstance

from ...testing import TestCase
from ..creds import Credentials


class TestCredentials(TestCase):
    """Tests for `maas.client.creds.Credentials`."""

    def test_str_form_is_colon_separated_triple(self):
        creds = Credentials("foo", "bar", "baz")
        self.assertEqual(":".join(creds), str(creds))

    def test_parse_reads_a_colon_separated_triple(self):
        creds = Credentials.parse("foo:bar:baz")
        self.assertEqual(("foo", "bar", "baz"), creds)
        self.assertThat(creds, IsInstance(Credentials))
        self.assertEqual("foo", creds.consumer_key)
        self.assertEqual("bar", creds.token_key)
        self.assertEqual("baz", creds.token_secret)

    def test_parse_reads_a_sequence(self):
        creds = Credentials.parse(("foo", "bar", "baz"))
        self.assertEqual(("foo", "bar", "baz"), creds)
        self.assertThat(creds, IsInstance(Credentials))
        self.assertEqual("foo", creds.consumer_key)
        self.assertEqual("bar", creds.token_key)
        self.assertEqual("baz", creds.token_secret)

    def test_parse_rejects_too_few_parts(self):
        self.assertRaises(ValueError, Credentials.parse, "foo:bar")
        self.assertRaises(ValueError, Credentials.parse, ("foo", "bar"))

    def test_parse_rejects_too_many_parts(self):
        self.assertRaises(ValueError, Credentials.parse, "a:b:c:d")
        self.assertRaises(ValueError, Credentials.parse, ("a", "b", "c", "d"))

    def test_parse_returns_None_when_there_are_no_parts(self):
        self.assertIsNone(Credentials.parse(""))
        self.assertIsNone(Credentials.parse(()))

    def test_parse_returns_None_when_passed_None(self):
        self.assertIsNone(Credentials.parse(None))

    def test_parse_returns_credentials_when_passed_credentials(self):
        creds = Credentials("foo", "bar", "baz")
        self.assertIs(creds, Credentials.parse(creds))



"""Commands for files."""

__all__ = ["register"]

from . import OriginPagedTableCommand, tables


class cmd_files(OriginPagedTableCommand):
    """List files."""

    def execute(self, origin, options, target):
        table = tables.FilesTable()
        return table.render(target, origin.Files.read())


def register(parser):
    """Register profile commands with the given parser."""
    cmd_files.register(parser)



"""Test for `maas.client.viscera.account`."""

from testtools.matchers import Equals, IsInstance, MatchesAll

from .. import account
from ...testing import make_name_without_spaces, TestCase
from ...utils import creds
from ...utils.testing import make_Credentials
from ..testing import bind


def make_origin():
    # Create a new origin with Account.
    return bind(account.Account)


class TestAccount(TestCase):
    def test__create_credentials_returns_Credentials(self):
        consumer_key = (make_name_without_spaces("consumer_key"),)
        token_key = (make_name_without_spaces("token_key"),)
        token_secret = (make_name_without_spaces("token_secret"),)

        origin = make_origin()
        create_authorisation_token = origin.Account._handler.create_authorisation_token
        create_authorisation_token.return_value = {
            "consumer_key": consumer_key,
            "token_key": token_key,
            "token_secret": token_secret,
        }

        credentials = origin.Account.create_credentials()
        self.assertThat(
            credentials,
            MatchesAll(
                IsInstance(creds.Credentials),
                Equals((consumer_key, token_key, token_secret)),
            ),
        )

    def test__create_credentials_ignores_other_keys_in_response(self):
        consumer_key = (make_name_without_spaces("consumer_key"),)
        token_key = (make_name_without_spaces("token_key"),)
        token_secret = (make_name_without_spaces("token_secret"),)

        origin = make_origin()
        create_authorisation_token = origin.Account._handler.create_authorisation_token
        create_authorisation_token.return_value = {
            "name": "cookie-monster",
            "fur-colour": "blue",
            "consumer_key": consumer_key,
            "token_key": token_key,
            "token_secret": token_secret,
        }

        credentials = origin.Account.create_credentials()
        self.assertThat(
            credentials,
            MatchesAll(
                IsInstance(creds.Credentials),
                Equals((consumer_key, token_key, token_secret)),
            ),
        )

    def test__delete_credentials_sends_token_key(self):
        origin = make_origin()
        delete_authorisation_token = origin.Account._handler.delete_authorisation_token
        delete_authorisation_token.return_value = None
        credentials = make_Credentials()

        origin.Account.delete_credentials(credentials)

        delete_authorisation_token.assert_called_once_with(
            token_key=credentials.token_key
        )



"""Test for `maas.client.viscera.bcache_cache_sets`."""

import random

from testtools.matchers import Equals

from ..bcache_cache_sets import BcacheCacheSet, BcacheCacheSets
from ..block_devices import BlockDevice, BlockDevices
from ..nodes import Node, Nodes
from ..partitions import Partition, Partitions

from ..testing import bind
from ...enum import BlockDeviceType
from ...testing import make_string_without_spaces, TestCase


def make_origin():
    """
    Create a new origin with required objects.
    """
    return bind(
        BcacheCacheSets,
        BcacheCacheSet,
        BlockDevices,
        BlockDevice,
        Partitions,
        Partition,
        Nodes,
        Node,
    )


class TestBcacheCacheSets(TestCase):
    def test__read_bad_node_type(self):
        BcacheCacheSets = make_origin().BcacheCacheSets
        error = self.assertRaises(
            TypeError, BcacheCacheSets.read, random.randint(0, 100)
        )
        self.assertEquals("node must be a Node or str, not int", str(error))

    def test__read_with_system_id(self):
        BcacheCacheSets = make_origin().BcacheCacheSets
        system_id = make_string_without_spaces()
        cache_sets = [
            {
                "system_id": system_id,
                "id": random.randint(0, 100),
                "cache_device": {
                    "id": random.randint(0, 100),
                    "type": BlockDeviceType.PHYSICAL.value,
                },
            }
            for _ in range(3)
        ]
        BcacheCacheSets._handler.read.return_value = cache_sets
        cache_sets = BcacheCacheSets.read(system_id)
        self.assertThat(len(cache_sets), Equals(3))
        BcacheCacheSets._handler.read.assert_called_once_with(system_id=system_id)

    def test__read_with_Node(self):
        origin = make_origin()
        BcacheCacheSets, Node = origin.BcacheCacheSets, origin.Node
        system_id = make_string_without_spaces()
        node = Node(system_id)
        cache_sets = [
            {
                "system_id": system_id,
                "id": random.randint(0, 100),
                "cache_device": {
                    "id": random.randint(0, 100),
                    "type": BlockDeviceType.PHYSICAL.value,
                },
            }
            for _ in range(3)
        ]
        BcacheCacheSets._handler.read.return_value = cache_sets
        cache_sets = BcacheCacheSets.read(node)
        self.assertThat(len(cache_sets), Equals(3))
        BcacheCacheSets._handler.read.assert_called_once_with(system_id=system_id)

    def test__create_bad_node_type(self):
        origin = make_origin()
        BcacheCacheSets = origin.BcacheCacheSets
        error = self.assertRaises(
            TypeError,
            BcacheCacheSets.create,
            random.randint(0, 100),
            origin.BlockDevice({}),
        )
        self.assertEquals("node must be a Node or str, not int", str(error))

    def test__create_bad_cache_device_type(self):
        BcacheCacheSets = make_origin().BcacheCacheSets
        error = self.assertRaises(
            TypeError,
            BcacheCacheSets.create,
            make_string_without_spaces(),
            random.randint(0, 100),
        )
        self.assertEquals(
            "cache_device must be a BlockDevice or Partition, not int", str(error)
        )

    def test__create_with_block_device(self):
        origin = make_origin()
        BcacheCacheSets = origin.BcacheCacheSets
        BlockDevice = origin.BlockDevice
        block_device = BlockDevice({"id": random.randint(0, 100)})
        BcacheCacheSets._handler.create.return_value = {
            "id": random.randint(0, 100),
            "cache_device": block_device,
        }
        system_id = make_string_without_spaces()
        BcacheCacheSets.create(system_id, block_device)
        BcacheCacheSets._handler.create.assert_called_once_with(
            system_id=system_id, cache_device=block_device.id
        )

    def test__create_with_partition(self):
        origin = make_origin()
        BcacheCacheSets = origin.BcacheCacheSets
        Partition = origin.Partition
        partition = Partition({"id": random.randint(0, 100)})
        BcacheCacheSets._handler.create.return_value = {
            "id": random.randint(0, 100),
            "cache_device": partition,
        }
        system_id = make_string_without_spaces()
        BcacheCacheSets.create(system_id, partition)
        BcacheCacheSets._handler.create.assert_called_once_with(
            system_id=system_id, cache_partition=partition.id
        )


class TestBcacheCacheSet(TestCase):
    def test__read_bad_node_type(self):
        BcacheCacheSet = make_origin().BcacheCacheSet
        error = self.assertRaises(
            TypeError,
            BcacheCacheSet.read,
            random.randint(0, 100),
            random.randint(0, 100),
        )
        self.assertEquals("node must be a Node or str, not int", str(error))

    def test__read_with_system_id(self):
        BcacheCacheSet = make_origin().BcacheCacheSet
        system_id = make_string_without_spaces()
        cache_set = {
            "system_id": system_id,
            "id": random.randint(0, 100),
            "cache_device": {
                "id": random.randint(0, 100),
                "type": BlockDeviceType.PHYSICAL.value,
            },
        }
        BcacheCacheSet._handler.read.return_value = cache_set
        BcacheCacheSet.read(system_id, cache_set["id"])
        BcacheCacheSet._handler.read.assert_called_once_with(
            system_id=system_id, id=cache_set["id"]
        )

    def test__read_with_Node(self):
        origin = make_origin()
        BcacheCacheSet = origin.BcacheCacheSet
        Node = origin.Node
        system_id = make_string_without_spaces()
        node = Node(system_id)
        cache_set = {
            "system_id": system_id,
            "id": random.randint(0, 100),
            "cache_device": {
                "id": random.randint(0, 100),
                "type": BlockDeviceType.PHYSICAL.value,
            },
        }
        BcacheCacheSet._handler.read.return_value = cache_set
        BcacheCacheSet.read(node, cache_set["id"])
        BcacheCacheSet._handler.read.assert_called_once_with(
            system_id=system_id, id=cache_set["id"]
        )

    def test__delete(self):
        BcacheCacheSet = make_origin().BcacheCacheSet
        system_id = make_string_without_spaces()
        cache_set = BcacheCacheSet(
            {
                "system_id": system_id,
                "id": random.randint(0, 100),
                "cache_device": {
                    "id": random.randint(0, 100),
                    "type": BlockDeviceType.PHYSICAL.value,
                },
            }
        )
        BcacheCacheSet._handler.delete.return_value = None
        cache_set.delete()
        BcacheCacheSet._handler.delete.assert_called_once_with(
            system_id=system_id, id=cache_set.id
        )



"""Objects for logical volumes."""

__all__ = ["LogicalVolume", "LogicalVolumes"]

from typing import Iterable

from .block_devices import (
    BlockDevice,
    BlockDevices,
    BlockDevicesType,
    BlockDeviceTypeMeta,
)
from .volume_groups import VolumeGroup
from ..utils import remove_None


class LogicalVolumesType(BlockDevicesType):
    """Metaclass for `LogicalVolumes`."""

    def bind(cls, origin, handler, handlers, *, name=None):
        # LogicalVolumes is just a wrapper over BlockDevices. So the
        # `BlockDevices` handler is binded instead of an empty handler.
        handler = handlers.get("BlockDevices")
        return super(LogicalVolumesType, cls).bind(origin, handler, handlers)

    async def create(
        cls,
        volume_group: VolumeGroup,
        name: str,
        size: int,
        *,
        uuid: str = None,
        tags: Iterable[str] = None
    ):
        """
        Create a logical volume on the volume group.

        :param volume_group: Volume group to create the logical volume on.
        :type node: `VolumeGroup`
        :param name: The name for the logical volume.
        :type name: `str`
        :param size: The size of the logical volume in bytes.
        :type size: `int`
        :param uuid: UUID of the logical volume.
        :type uuid: `str`
        :param tags: List of tags to add to the logical volume.
        :type tags: sequence of `str`
        """
        if not isinstance(volume_group, VolumeGroup):
            raise TypeError(
                "volume_group must be a VolumeGroup, not %s"
                % (type(volume_group).__name__)
            )

        params = {"system_id": volume_group.node.system_id, "id": volume_group.id}
        if not name:
            raise ValueError("name must be provided.")
        if not size or size < 0:
            raise ValueError("size must be provided and greater than zero.")

        params.update(remove_None({"name": name, "size": size, "uuid": uuid}))
        data = await volume_group._handler.create_logical_volume(**params)
        # Create logical volume doesn't return a full block device object.
        # Load the logical volume using the block device endpoint, ensures that
        # all the data present to access the fields.
        bd_handler = getattr(cls._origin, "BlockDevice")._handler
        volume = cls._object(
            await bd_handler.read(system_id=data["system_id"], id=data["id"])
        )
        if tags:
            volume.tags = tags
            await volume.save()
        return volume


class LogicalVolumes(BlockDevices, metaclass=LogicalVolumesType):
    """The set of logical volumes on a volume group."""


class LogicalVolumeType(BlockDeviceTypeMeta):
    """Metaclass for `LogicalVolume`."""

    def bind(cls, origin, handler, handlers, *, name=None):
        # LogicalVolume is just a wrapper over BlockDevice. So the
        # `BlockDevice` handler is binded instead of an empty handler.
        handler = handlers.get("BlockDevice")
        return super(LogicalVolumeType, cls).bind(origin, handler, handlers)


class LogicalVolume(BlockDevice, metaclass=LogicalVolumeType):
    """A logical volume on a volume group."""

    def __repr__(self):
        return super(BlockDevice, self).__repr__(name="LogicalVolume", fields={"name"})



"""Objects for tags."""

__all__ = ["Tag", "Tags"]

from . import check, check_optional, Object, ObjectField, ObjectSet, ObjectType
from .nodes import Node


class TagsType(ObjectType):
    """Metaclass for `Tags`."""

    async def create(cls, name, *, comment=None, definition=None, kernel_opts=None):
        params = {"name": name}
        if comment is not None:
            params["comment"] = comment
        if definition is not None:
            params["definition"] = definition
        if kernel_opts is not None:
            params["kernel_opts"] = kernel_opts
        data = await cls._handler.create(**params)
        return cls._object(data)

    async def read(cls):
        data = await cls._handler.read()
        return cls(map(cls._object, data))


class Tags(ObjectSet, metaclass=TagsType):
    """The set of tags."""

    @classmethod
    def Managed(cls, manager, field, items):
        """Create a custom `Tags` that is managed by a related `Node.`

        :param manager: The manager of the `ObjectSet`. This is the `Object`
            that manages this set of objects.
        :param field: The field on the `manager` that created this managed
            `ObjectSet`.
        :param items: The items in the `ObjectSet`.
        """
        if not isinstance(manager, Node):
            raise TypeError(
                "manager must be instance of Node, not %s", type(manager).__name__
            )

        async def add(self, tag: Tag):
            """Add `tag` to node.

            :param tag: Tag to add to the node.
            :type tag: `Tag`.
            """
            if not isinstance(tag, Tag):
                raise TypeError(
                    "tag must be instance of Tag, not %s", type(tag).__name__
                )
            await tag._handler.update_nodes(name=tag.name, add=manager.system_id)
            if tag.name not in manager._data[field.name]:
                manager._data[field.name] += [tag.name]

        async def remove(self, tag: Tag):
            """Remove `tag` from node.

            :param tag: Tag to from the node.
            :type tag: `Tag`.
            """
            if not isinstance(tag, Tag):
                raise TypeError(
                    "tag must be instance of Tag, not %s", type(tag).__name__
                )
            await tag._handler.update_nodes(name=tag.name, remove=manager.system_id)
            manager._data[field.name] = [
                tag_name
                for tag_name in manager._data[field.name]
                if tag_name != tag.name
            ]

        attrs = {"add": add, "remove": remove}
        cls = type(
            "%s.Managed#%s" % (cls.__name__, manager.__class__.__name__), (cls,), attrs
        )
        return cls(items)


class TagType(ObjectType):
    async def read(cls, name):
        data = await cls._handler.read(name=name)
        return cls(data)


class Tag(Object, metaclass=TagType):
    """A tag."""

    name = ObjectField.Checked("name", check(str), readonly=True, pk=True)
    comment = ObjectField.Checked("comment", check(str), check(str), default="")
    definition = ObjectField.Checked("definition", check(str), check(str), default="")
    kernel_opts = ObjectField.Checked(
        "kernel_opts", check_optional(str), check_optional(str), default=None
    )

    async def delete(self):
        """
        Deletes the `Tag` from MAAS.
        """
        return await self._handler.delete(name=self.name)



"""Tests for `maas.client.flesh.controllers`."""

from operator import itemgetter
import yaml

from .testing import TestCaseWithProfile
from .. import ArgumentParser, controllers, tabular
from ...enum import NodeType
from ...testing import make_name_without_spaces
from ...viscera.testing import bind
from ...viscera.controllers import (
    RackController,
    RackControllers,
    RegionController,
    RegionControllers,
)


def make_origin():
    """Make origin for controllers."""
    return bind(RackControllers, RackController, RegionController, RegionControllers)


class TestControllers(TestCaseWithProfile):
    """Tests for `cmd_controllers`."""

    def test_returns_table_with_controllers(self):
        origin = make_origin()
        parser = ArgumentParser()
        region_rack_id = make_name_without_spaces()
        region_rack_hostname = make_name_without_spaces()
        racks = [
            {
                "system_id": region_rack_id,
                "hostname": region_rack_hostname,
                "node_type": NodeType.REGION_AND_RACK_CONTROLLER.value,
                "architecture": "amd64/generic",
                "cpu_count": 2,
                "memory": 1024,
            },
            {
                "system_id": make_name_without_spaces(),
                "hostname": make_name_without_spaces(),
                "node_type": NodeType.RACK_CONTROLLER.value,
                "architecture": "amd64/generic",
                "cpu_count": 2,
                "memory": 1024,
            },
        ]
        regions = [
            {
                "system_id": region_rack_id,
                "hostname": region_rack_hostname,
                "node_type": NodeType.REGION_AND_RACK_CONTROLLER.value,
                "architecture": "amd64/generic",
                "cpu_count": 2,
                "memory": 1024,
            },
            {
                "system_id": make_name_without_spaces(),
                "hostname": make_name_without_spaces(),
                "node_type": NodeType.REGION_CONTROLLER.value,
                "architecture": "amd64/generic",
                "cpu_count": 2,
                "memory": 1024,
            },
        ]
        origin.RackControllers._handler.read.return_value = racks
        origin.RegionControllers._handler.read.return_value = regions
        cmd = controllers.cmd_controllers(parser)
        subparser = controllers.cmd_controllers.register(parser)
        options = subparser.parse_args([])
        output = yaml.safe_load(
            cmd.execute(origin, options, target=tabular.RenderTarget.yaml)
        )
        self.assertEquals(
            [
                {"name": "hostname", "title": "Hostname"},
                {"name": "node_type", "title": "Type"},
                {"name": "architecture", "title": "Arch"},
                {"name": "cpus", "title": "#CPUs"},
                {"name": "memory", "title": "RAM"},
            ],
            output["columns"],
        )
        controller_output = {
            controller["hostname"]: {
                "hostname": controller["hostname"],
                "node_type": controller["node_type"],
                "architecture": controller["architecture"],
                "cpus": controller["cpu_count"],
                "memory": controller["memory"],
            }
            for controller in racks + regions
        }
        self.assertEquals(
            sorted(controller_output.values(), key=itemgetter("hostname")),
            output["data"],
        )

    def test_calls_handler_with_hostnames(self):
        origin = make_origin()
        parser = ArgumentParser()
        origin.RackControllers._handler.read.return_value = []
        origin.RegionControllers._handler.read.return_value = []
        subparser = controllers.cmd_controllers.register(parser)
        cmd = controllers.cmd_controllers(parser)
        hostnames = [make_name_without_spaces() for _ in range(3)]
        options = subparser.parse_args(hostnames)
        cmd.execute(origin, options, target=tabular.RenderTarget.yaml)
        origin.RackControllers._handler.read.assert_called_once_with(hostname=hostnames)
        origin.RegionControllers._handler.read.assert_called_once_with(
            hostname=hostnames
        )



"""Test helpers for `maas.client.flesh`."""

__all__ = ["capture_parse_error"]

import argparse

from .... import flesh
from ....testing import TestCase
from ....utils.tests.test_profiles import make_profile


def capture_parse_error(parser, *args):
    """Capture the `ArgumentError` arising from parsing the given arguments.

    `argparse` is hard to test (and to introspect, and extend... but it is
    good at what it does) so we have to use a pseudo-private method here.
    """
    namespace = argparse.Namespace()
    try:
        parser._parse_known_args(list(args), namespace)
    except argparse.ArgumentError as error:
        return error
    else:
        return None


class TestCaseWithProfile(TestCase):
    """Base test case class for all of `flesh` commands.

    This creates an empty default profile.
    """

    def setUp(self):
        self.profile = make_profile("default")
        self.patch(flesh, "PROFILE_NAMES", ["default"])
        self.patch(flesh, "PROFILE_DEFAULT", self.profile)
        super(TestCaseWithProfile, self).setUp()



"""Test for `maas.client.viscera.files`."""

from testtools.matchers import AllMatch, IsInstance, MatchesSetwise, MatchesStructure

from .. import files
from ...testing import make_name_without_spaces, TestCase
from ..testing import bind


def make_origin():
    # Create a new origin with Files and File. The former refers to the
    # latter via the origin, hence why it must be bound.
    origin = bind(files.Files, files.File)
    return origin


class TestFiles(TestCase):
    """Tests for `Files`."""

    def test__read(self):
        origin = make_origin()

        data = [
            {"filename": make_name_without_spaces()},
            {"filename": make_name_without_spaces()},
        ]
        origin.Files._handler.read.return_value = data

        resources = origin.Files.read()
        self.assertEquals(2, len(resources))
        self.assertThat(resources, IsInstance(origin.Files))
        self.assertThat(resources, AllMatch(IsInstance(origin.File)))
        self.assertThat(
            resources,
            MatchesSetwise(
                *(
                    MatchesStructure.byEquality(filename=entry["filename"])
                    for entry in data
                )
            ),
        )


class TestFile(TestCase):
    """Tests for `File`."""

    def test__read(self):
        origin = make_origin()
        data = {"filename": make_name_without_spaces()}
        self.assertThat(
            origin.File(data), MatchesStructure.byEquality(filename=data["filename"])
        )



"""Test for `maas.client.viscera.controllers`."""

from testtools.matchers import Equals

from .. import controllers
from ...testing import make_name_without_spaces, TestCase
from ..testing import bind


def make_origin():
    # Create a new origin with RackController, RegionControllers,
    # RackController, and RegionController.
    return bind(
        controllers.RackControllers,
        controllers.RackController,
        controllers.RegionControllers,
        controllers.RegionController,
    )


class TestRackController(TestCase):
    def test__string_representation_includes_only_system_id_and_hostname(self):
        rack_controller = controllers.RackController(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
            }
        )
        self.assertThat(
            repr(rack_controller),
            Equals(
                "<RackController hostname=%(hostname)r system_id=%(system_id)r>"
                % rack_controller._data
            ),
        )

    def test__get_power_parameters(self):
        rack_controller = make_origin().RackController(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
            }
        )
        power_parameters = {"key": make_name_without_spaces("value")}
        rack_controller._handler.power_parameters.return_value = power_parameters
        self.assertThat(
            rack_controller.get_power_parameters(), Equals(power_parameters)
        )
        rack_controller._handler.power_parameters.assert_called_once_with(
            system_id=rack_controller.system_id
        )

    def test__set_power(self):
        orig_power_type = make_name_without_spaces("power_type")
        new_power_type = make_name_without_spaces("power_type")
        rack_controller = make_origin().RackController(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
                "power_type": orig_power_type,
            }
        )
        power_parameters = {"key": make_name_without_spaces("value")}
        rack_controller._handler.update.return_value = {"power_type": new_power_type}
        rack_controller.set_power(new_power_type, power_parameters)
        rack_controller._handler.update.assert_called_once_with(
            system_id=rack_controller.system_id,
            power_type=new_power_type,
            power_parameters=power_parameters,
        )
        self.assertThat(rack_controller.power_type, Equals(new_power_type))


class TestRegionController(TestCase):
    def test__string_representation_includes_only_system_id_and_hostname(self):
        rack_controller = controllers.RegionController(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
            }
        )
        self.assertThat(
            repr(rack_controller),
            Equals(
                "<RegionController hostname=%(hostname)r system_id=%(system_id)r>"
                % rack_controller._data
            ),
        )

    def test__get_power_parameters(self):
        region_controller = make_origin().RegionController(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
            }
        )
        power_parameters = {"key": make_name_without_spaces("value")}
        region_controller._handler.power_parameters.return_value = power_parameters
        self.assertThat(
            region_controller.get_power_parameters(), Equals(power_parameters)
        )
        region_controller._handler.power_parameters.assert_called_once_with(
            system_id=region_controller.system_id
        )

    def test__set_power(self):
        orig_power_type = make_name_without_spaces("power_type")
        new_power_type = make_name_without_spaces("power_type")
        region_controller = make_origin().RegionController(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
                "power_type": orig_power_type,
            }
        )
        power_parameters = {"key": make_name_without_spaces("value")}
        region_controller._handler.update.return_value = {"power_type": new_power_type}
        region_controller.set_power(new_power_type, power_parameters)
        region_controller._handler.update.assert_called_once_with(
            system_id=region_controller.system_id,
            power_type=new_power_type,
            power_parameters=power_parameters,
        )
        self.assertThat(region_controller.power_type, Equals(new_power_type))



"""Setuptools installer for python-libmaas."""

from os.path import (
    dirname,
    join,
)

from setuptools import (
    find_packages,
    setup,
)


# The directory in which setup.py lives.
here = dirname(__file__)


def read(filename):
    """Return the whitespace-stripped content of `filename`."""
    path = join(here, filename)
    with open(path, "r") as fin:
        return fin.read().strip()


setup(
    name='python-libmaas',
    author='MAAS Developers',
    author_email='maas-devel@lists.launchpad.net',
    url='https://github.com/maas/python-libmaas',
    version="0.6.8",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Software Development :: Libraries',
    ],
    namespace_packages=['maas'],
    packages=find_packages(
        include={"maas", "maas.*"},
        exclude={"*.tests", "*.testing"},
    ),
    package_data={
        'maas.client.bones.testing': ['*.json'],
    },
    install_requires=[
        "aiohttp >= 2.0.0",
        "argcomplete >= 1.0",
        "colorclass >= 1.2.0",
        "macaroonbakery >= 1.1.3",
        "oauthlib >= 1.0.3",
        "pymongo >= 3.5.1",  # for bson
        "pytz >= 2014.10",
        "PyYAML >= 3.11",
        "terminaltables >= 2.1.0",
    ],
    test_suite="maas.client",
    tests_require=[
        "django >= 2.2.4, < 3.0",
        "fixtures >= 1.0.0",
        "setuptools",
        "testscenarios",
        "testtools",
        "Twisted",
    ],
    description="A client API library specially for MAAS.",
    long_description=read('README'),
    long_description_content_type='text/markdown',
    entry_points={
        "console_scripts": {
            "maas = maas.client.flesh:main",
        },
    },
)



"""Commands for interacting with a remote MAAS."""

__all__ = [
    "colorized",
    "Command",
    "CommandError",
    "OriginCommand",
    "OriginTableCommand",
    "PROFILE_DEFAULT",
    "PROFILE_NAMES",
    "TableCommand",
]

from abc import ABCMeta, abstractmethod
import argparse
from importlib import import_module
import os
import subprocess
import sys
import textwrap
import typing

import argcomplete
import colorclass

from . import tabular
from .. import bones, utils, viscera
from ..utils.auth import try_getpass
from ..utils.profiles import Profile, ProfileStore


PROG_DESCRIPTION = """\
MAAS provides complete automation of your physical servers for amazing data
center operational efficiency.

See https://maas.io/docs for documentation.

Common commands:

    {program} login           Log-in to a MAAS.
    {program} switch          Switch the active profile.
    {program} machines        List machines.
    {program} deploy          Allocate and deploy machine.
    {program} release         Release machine.
    {program} fabrics         List fabrics.
    {program} subnets         List subnets.

Example help commands:

    `{program} help`          This help page
    `{program} help commands` Lists all commands
    `{program} help deploy`   Shows help for command 'deploy'
"""


def colorized(text):
    if sys.stdout.isatty():
        # Don't return value_colors; returning the Color instance allows
        # terminaltables to correctly calculate alignment and padding.
        return colorclass.Color(text)
    else:
        return colorclass.Color(text).value_no_colors


def read_input(message, validator=None, password=False):
    message = "%s: " % message
    while True:
        if password:
            value = try_getpass(message)
        else:
            value = input(message)
        if value:
            if validator is not None:
                try:
                    validator(value)
                except Exception as exc:
                    print(colorized("{{autored}}Error: {{/autored}} %s") % str(exc))
                else:
                    return value
            else:
                return value


def yes_or_no(question):
    question = "%s [y/N] " % question
    while True:
        value = input(question)
        value = value.lower()
        if value in ["y", "yes"]:
            return True
        elif value in ["n", "no"]:
            return False


def print_with_pager(output):
    """Print the output to `stdout` using less when in a tty."""
    if sys.stdout.isatty():
        try:
            pager = subprocess.Popen(
                ["less", "-F", "-r", "-S", "-X", "-K"],
                stdin=subprocess.PIPE,
                stdout=sys.stdout,
            )
        except subprocess.CalledProcessError:
            # Don't use the pager since starting it has failed.
            print(output)
            return
        else:
            pager.stdin.write(output.encode("utf-8"))
            pager.stdin.close()
            pager.wait()
    else:
        # Output directly to stdout since not in tty.
        print(output)


def get_profile_names_and_default() -> (
    typing.Tuple[typing.Sequence[str], typing.Optional[Profile]]
):
    """Return the list of profile names and the default profile object.

    The list of names is sorted.
    """
    with ProfileStore.open() as config:
        return sorted(config), config.default


# Get profile names and the default profile now to avoid repetition when
# defining arguments (e.g. default and choices). Doing this as module-import
# time is imperfect but good enough for now.
PROFILE_NAMES, PROFILE_DEFAULT = get_profile_names_and_default()


class MinimalHelpAction(argparse._HelpAction):
    def __call__(self, parser, namespace, values, option_string=None):
        parser.print_minized_help()
        parser.exit()


class PagedHelpAction(argparse._HelpAction):
    def __call__(self, parser, namespace, values, option_string=None):
        print_with_pager(parser.format_help())
        parser.exit()


class HelpFormatter(argparse.RawDescriptionHelpFormatter):
    """Specialization of argparse's raw description help formatter to modify
    usage to be in a better format.
    """

    def _format_usage(self, usage, actions, groups, prefix):
        if prefix is None:
            prefix = "Usage: "
        return super(HelpFormatter, self)._format_usage(usage, actions, groups, prefix)


class ArgumentParser(argparse.ArgumentParser):
    """Specialization of argparse's parser with better support for
    subparsers and better help output.

    Specifically, the one-shot `add_subparsers` call is disabled, replaced by
    a lazily evaluated `subparsers` property.

    `print_minized_help` is added to only show the description which is
    specially formatted.
    """

    def add_subparsers(self):
        raise NotImplementedError("add_subparsers has been disabled")

    @property
    def subparsers(self):
        """Obtain the subparser's object."""
        try:
            return self.__subparsers
        except AttributeError:
            parent = super(ArgumentParser, self)
            self.__subparsers = parent.add_subparsers(title="drill down")
            self.__subparsers.metavar = "COMMAND"
            return self.__subparsers

    def add_argument_group(self, title, description=None):
        """Add an argument group, or return a pre-existing one."""
        try:
            groups = self.__groups
        except AttributeError:
            groups = self.__groups = {}

        if title not in groups:
            groups[title] = super().add_argument_group(
                title=title, description=description
            )

        return groups[title]

    @property
    def other(self):
        return self.add_argument_group("other arguments")

    def __getitem__(self, name):
        """Return the named subparser."""
        return self.subparsers.choices[name]

    def error(self, message):
        """Make the default error messages more helpful

        Override default ArgumentParser error method to print the help menu
        generated by ArgumentParser instead of just printing out a list of
        valid arguments.
        """
        self.exit(2, colorized("{autored}Error:{/autored} ") + message + "\n")

    def print_minized_help(self, *, no_pager=False):
        """Return the formatted help text.

        Override default ArgumentParser to just include the usage and the
        description. The `help` action is used for provide more detail.
        """
        formatter = self._get_formatter()
        formatter.add_usage(self.usage, self._actions, self._mutually_exclusive_groups)
        formatter.add_text(self.description)
        if no_pager:
            print(formatter.format_help())
        else:
            print_with_pager(formatter.format_help())


class CommandError(Exception):
    """A command has failed during execution."""


class Command(metaclass=ABCMeta):
    """A base class for composing commands.

    This adheres to the expectations of `register`.
    """

    def __init__(self, parser):
        super(Command, self).__init__()
        self.parser = parser

    @abstractmethod
    def __call__(self, options):
        """Execute this command."""

    @classmethod
    def name(cls):
        """Return the preferred name as which this command will be known."""
        name = cls.__name__.replace("_", "-").lower()
        name = name[4:] if name.startswith("cmd-") else name
        return name

    @classmethod
    def register(cls, parser, name=None):
        """Register this command as a sub-parser of `parser`.

        :type parser: An instance of `ArgumentParser`.
        :return: The sub-parser created.
        """
        help_title, help_body = utils.parse_docstring(cls)
        command_parser = parser.subparsers.add_parser(
            cls.name() if name is None else name,
            help=help_title,
            description=help_title,
            epilog=help_body,
            add_help=False,
            formatter_class=HelpFormatter,
        )
        command_parser.add_argument(
            "-h", "--help", action=PagedHelpAction, help=argparse.SUPPRESS
        )
        command_parser.set_defaults(execute=cls(command_parser))
        return command_parser


class TableCommand(Command):
    def __init__(self, parser):
        super(TableCommand, self).__init__(parser)
        if sys.stdout.isatty():
            default_target = tabular.RenderTarget.pretty
        else:
            default_target = tabular.RenderTarget.plain
        parser.other.add_argument(
            "--format",
            type=tabular.RenderTarget,
            choices=tabular.RenderTarget,
            default=default_target,
            help=(
                "Output tabular data as a formatted table (pretty), a "
                "formatted table using only ASCII for borders (plain), or "
                "one of several dump formats. Default: %(default)s."
            ),
        )


class OriginCommandBase(Command):
    def __init__(self, parser):
        super(OriginCommandBase, self).__init__(parser)
        parser.other.add_argument(
            "--profile",
            dest="profile_name",
            metavar="NAME",
            choices=PROFILE_NAMES,
            required=(PROFILE_DEFAULT is None),
            help=(
                "The name of the remote MAAS instance to use. Use "
                "`profiles list` to obtain a list of valid profiles."
                + (
                    ""
                    if PROFILE_DEFAULT is None
                    else " [default: %s]" % PROFILE_DEFAULT.name
                )
            ),
        )
        if PROFILE_DEFAULT is not None:
            parser.set_defaults(profile=PROFILE_DEFAULT.name)


class OriginCommand(OriginCommandBase):
    def __call__(self, options):
        session = bones.SessionAPI.fromProfileName(options.profile)
        origin = viscera.Origin(session)
        return self.execute(origin, options)

    def execute(self, origin, options):
        raise NotImplementedError("Implement execute() in subclasses.")


class OriginTableCommand(OriginCommandBase, TableCommand):
    def __call__(self, options):
        session = bones.SessionAPI.fromProfileName(options.profile)
        origin = viscera.Origin(session)
        return self.execute(origin, options, target=options.format)

    def execute(self, origin, options, *, target):
        raise NotImplementedError("Implement execute() in subclasses.")


class OriginPagedTableCommand(OriginTableCommand):
    def __init__(self, parser):
        super(OriginPagedTableCommand, self).__init__(parser)
        parser.other.add_argument(
            "--no-pager",
            action="store_true",
            help=("Don't use the pager when printing the output of the " "command."),
        )

    def __call__(self, options):
        return_code = 0
        output = super(OriginPagedTableCommand, self).__call__(options)
        if isinstance(output, tuple):
            return_code, output = output
        elif isinstance(output, int):
            return_code = output
            output = None
        elif isinstance(output, str):
            pass
        else:
            raise TypeError(
                "execute must return either tuple, int or str, not %s"
                % (type(output).__name__)
            )
        if output:
            if options.no_pager:
                print(output)
            else:
                print_with_pager(output)
        return return_code


class cmd_help(Command):
    """Show the help summary or help for a specific command."""

    def __init__(self, parser, parent_parser):
        self.parent_parser = parent_parser
        super(cmd_help, self).__init__(parser)
        parser.add_argument(
            "-h", "--help", action=PagedHelpAction, help=argparse.SUPPRESS
        )
        parser.add_argument("command", nargs="?", help="Show help for this command.")
        parser.other.add_argument(
            "--no-pager",
            action="store_true",
            help=("Don't use the pager when printing the output of the " "command."),
        )

    def __call__(self, options):
        if options.command is None:
            self.parent_parser.print_minized_help(no_pager=options.no_pager)
        else:
            command = self.parent_parser.subparsers.choices.get(options.command, None)
            if command is None:
                if options.command == "commands":
                    self.print_all_commands(no_pager=options.no_pager)
                else:
                    self.parser.error("unknown command %s" % options.command)
            else:
                if options.no_pager:
                    command.print_help()
                else:
                    print_with_pager(command.format_help())

    def print_all_commands(self, *, no_pager=False):
        """Print help for all commands.

        Commands are sorted in alphabetical order and wrapping is done
        based on the width of the terminal.
        """
        formatter = self.parent_parser._get_formatter()
        command_names = sorted(self.parent_parser.subparsers.choices.keys())
        max_name_len = max([len(name) for name in command_names]) + 1
        commands = ""
        for name in command_names:
            command = self.parent_parser.subparsers.choices[name]
            extra_padding = max_name_len - len(name)
            command_line = "%s%s%s" % (name, " " * extra_padding, command.description)
            while len(command_line) > formatter._width:
                lines = textwrap.wrap(command_line, formatter._width)
                commands += "%s\n" % lines[0]
                if len(lines) > 1:
                    lines[1] = (" " * max_name_len) + lines[1]
                    command_line = " ".join(lines[1:])
                else:
                    command_line = None
            if command_line:
                commands += "%s\n" % command_line
        if no_pager:
            print(commands[:-1])
        else:
            print_with_pager(commands[:-1])

    @classmethod
    def register(cls, parser, name=None):
        """Register this command as a sub-parser of `parser`.

        :type parser: An instance of `ArgumentParser`.
        :return: The sub-parser created.
        """
        help_title, help_body = utils.parse_docstring(cls)
        command_parser = parser.subparsers.add_parser(
            cls.name() if name is None else name,
            help=help_title,
            description=help_title,
            epilog=help_body,
            add_help=False,
            formatter_class=HelpFormatter,
        )
        command_parser.set_defaults(execute=cls(command_parser, parser))
        return command_parser


def prepare_parser(program):
    """Create and populate an argument parser."""
    parser = ArgumentParser(
        description=PROG_DESCRIPTION.format(program=program),
        prog=program,
        formatter_class=HelpFormatter,
        add_help=False,
    )
    parser.add_argument(
        "-h", "--help", action=MinimalHelpAction, help=argparse.SUPPRESS
    )

    # Register sub-commands.
    submodules = (
        "nodes",
        "machines",
        "devices",
        "controllers",
        "fabrics",
        "vlans",
        "subnets",
        "spaces",
        "files",
        "tags",
        "users",
        "profiles",
        "shell",
    )
    cmd_help.register(parser)
    for submodule in submodules:
        module = import_module("." + submodule, __name__)
        module.register(parser)

    # Register global options.
    parser.add_argument(
        "--debug", action="store_true", default=False, help=argparse.SUPPRESS
    )

    return parser


def post_mortem(traceback):
    """Work with an exception in a post-mortem debugger.

    Try to use `ipdb` first, falling back to `pdb`.
    """
    try:
        from ipdb import post_mortem
    except ImportError:
        from pdb import post_mortem

    message = "Entering post-mortem debugger. Type `help` for help."
    redline = colorized("{autored}%s{/autored}") % "{0:=^{1}}"

    print()
    print(redline.format(" CRASH! ", len(message)))
    print(message)
    print(redline.format("", len(message)))
    print()

    post_mortem(traceback)


def program_name_from_env(program):
    """Return the program name from environment."""
    if os.environ.get("SNAP_INSTANCE_NAME"):
        return os.environ.get("SNAP_INSTANCE_NAME")
    return program


def main(argv=sys.argv):
    program, *arguments = argv
    program = program_name_from_env(program)
    parser, options = None, None

    try:
        parser = prepare_parser(program)
        argcomplete.autocomplete(parser, exclude=("-h", "--help"))
        options = parser.parse_args(arguments)
        try:
            execute = options.execute
        except AttributeError:
            parser.error("Argument missing.")
        else:
            return execute(options)
    except KeyboardInterrupt:
        raise SystemExit(1)
    except Exception as error:
        # This is unexpected. Why? Because the CLI code raises SystemExit or
        # invokes something that raises SystemExit when it chooses to exit.
        # SystemExit does not subclass Exception, and so it would not be
        # handled here, hence this is not a deliberate exit.
        if parser is None or options is None or options.debug:
            # The user has either chosen to debug OR we crashed before/while
            # parsing arguments. Either way, let's not be terse.
            if sys.stdin.isatty() and sys.stdout.isatty():
                # We're at a fully interactive terminal so let's post-mortem.
                *_, exc_traceback = sys.exc_info()
                post_mortem(exc_traceback)
                # Exit non-zero, but quietly; dumping the traceback again on
                # the way out is confusing after doing a post-mortem.
                raise SystemExit(1)
            else:
                # Re-raise so the traceback is dumped and we exit non-zero.
                raise
        else:
            # Display a terse error message. Note that parser.error() will
            # raise SystemExit(>0) after printing its message.
            parser.error("%s" % error)



"""Test for `maas.client.viscera.version`."""

from random import randrange

from testtools.matchers import Equals

from .. import version
from ...testing import make_name_without_spaces, TestCase


class TestVersion(TestCase):
    def test__fields(self):
        v_major = randrange(2, 99)
        v_point1 = randrange(0, 99)
        v_point2 = randrange(0, 99)
        v_tuple = v_major, v_point1, v_point2
        v_string = "%d.%d.%d" % v_tuple
        v_sub = make_name_without_spaces("subv")
        v_caps = [make_name_without_spaces("cap") for _ in range(randrange(10))]

        vsn = version.Version(
            {"version": v_string, "subversion": v_sub, "capabilities": v_caps}
        )

        self.assertThat(vsn.version, Equals(v_string))
        self.assertThat(vsn.version_info, Equals(v_tuple))
        self.assertThat(vsn.subversion, Equals(v_sub))
        self.assertThat(vsn.capabilities, Equals(frozenset(v_caps)))

    def test__string_representation(self):
        vsn = version.Version(
            {
                "version": "2.1.0",
                "subversion": "alpha9999+bzr8888",
                "capabilities": ["kung-fu", "badgering"],
            }
        )
        self.assertThat(
            repr(vsn), Equals("<Version 2.1.0 alpha9999+bzr8888 [badgering kung-fu]>")
        )



"""Testing helpers for the Bones API."""

__all__ = ["api_descriptions", "DescriptionServer", "list_api_descriptions"]

import http
import http.server
import json
from operator import itemgetter
from pathlib import Path
import re
import threading

import fixtures
from pkg_resources import resource_filename, resource_listdir


def list_api_descriptions():
    """List API description documents.

    They're searched for in the same directory as this file, and their name
    must match "apiXX.json" where "XX" denotes the major and minor version
    number of the API.
    """
    for filename in resource_listdir(__name__, "."):
        match = re.match(r"api(\d)(\d)[.]json", filename)
        if match is not None:
            version = tuple(map(int, match.groups()))
            path = resource_filename(__name__, filename)
            name = "%d.%d" % version
            yield name, version, Path(path)


def load_api_descriptions():
    """Load the API description documents found by `list_api_descriptions`."""
    for name, version, path in list_api_descriptions():
        description = path.read_text("utf-8")
        yield name, version, json.loads(description)


api_descriptions = sorted(load_api_descriptions(), key=itemgetter(1))
assert len(api_descriptions) != 0


class DescriptionHandler(http.server.BaseHTTPRequestHandler):
    """An HTTP request handler that serves only API descriptions.

    The `desc` attribute ought to be specified, for example by subclassing, or
    by using the `make` class-method.

    The `content_type` attribute can be overridden to simulate a different
    Content-Type header for the description.
    """

    # Override these in subclasses.
    description = b'{"resources": []}'
    content_type = "application/json"

    @classmethod
    def make(cls, description=description):
        return type("DescriptionHandler", (cls,), {"description": description})

    def setup(self):
        super(DescriptionHandler, self).setup()
        self.logs = []

    def log_message(self, *args):
        """By default logs go to stdout/stderr. Instead, capture them."""
        self.logs.append(args)

    def do_GET(self):
        version_match = re.match(r"/MAAS/api/([0-9.]+)/describe/$", self.path)
        if version_match is None:
            self.send_error(http.HTTPStatus.NOT_FOUND)
        else:
            self.send_response(http.HTTPStatus.OK)
            self.send_header("Content-Type", self.content_type)
            self.send_header("Content-Length", str(len(self.description)))
            self.end_headers()
            self.wfile.write(self.description)


class DescriptionServer(fixtures.Fixture):
    """Fixture to start up an HTTP server for API descriptions only.

    :ivar handler: A `DescriptionHandler` subclass.
    :ivar server: An `http.server.HTTPServer` instance.
    :ivar url: A URL that points to the API that `server` is mocking.
    """

    def __init__(self, description=DescriptionHandler.description):
        super(DescriptionServer, self).__init__()
        self.description = description

    def _setUp(self):
        self.handler = DescriptionHandler.make(self.description)
        self.server = http.server.HTTPServer(("", 0), self.handler)
        self.url = "http://%s:%d/MAAS/api/2.0/" % self.server.server_address
        threading.Thread(target=self.server.serve_forever).start()
        self.addCleanup(self.server.server_close)
        self.addCleanup(self.server.shutdown)
